# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path
import csv

from pyomo.dataportal import TableData
from pyomo.dataportal.factory import DataManagerFactory


@DataManagerFactory.register("csv", "CSV file interface")
class CSVTable(TableData):
    def __init__(self):
        TableData.__init__(self)

    def open(self):
        if self.filename is None:  # pragma:nocover
            raise IOError("No filename specified")

    def close(self):
        self.FILE.close()

    def read(self):
        from pyomo.core.base.param import Param

        if not os.path.exists(self.filename):  # pragma:nocover
            raise IOError("Cannot find file '%s'" % self.filename)
        self.FILE = open(self.filename, 'r')
        tmp = []
        for tokens in csv.reader(self.FILE):
            if tokens != ['']:
                tmp.append(tokens)
        self.FILE.close()
        if len(tmp) == 0:
            raise IOError("Empty *.csv file")
        elif len(tmp) == 1:
            if not self.options.param is None:
                if type(self.options.param) in (list, tuple):
                    p = self.options.param[0]
                else:
                    p = self.options.param
                if isinstance(p, Param):
                    self.options.model = p.model()
                    p = p.local_name
                self._info = ["param", p, ":=", tmp[0][0]]
            elif len(self.options.symbol_map) == 1:
                self._info = [
                    "param",
                    self.options.symbol_map[self.options.symbol_map.keys()[0]],
                    ":=",
                    tmp[0][0],
                ]
            else:
                raise IOError(
                    "Data looks like a parameter, but multiple parameter names have been specified: %s"
                    % str(self.options.symbol_map)
                )
        else:
            self._set_data(tmp[0], tmp[1:])

    def write(self, data):
        if self.options.set is None and self.options.param is None:
            raise IOError("Unspecified model component")
        self.FILE = open(self.filename, 'w')
        table = self._get_table()
        writer = csv.writer(self.FILE)
        writer.writerows(table)
        self.FILE.close()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.numvalue import is_numeric_data
from pyomo.core.kernel.base import ICategorizedObject, _abstract_readwrite_property
from pyomo.core.kernel.container_utils import define_simple_containers


class ISOS(ICategorizedObject):
    """
    The interface for Special Ordered Sets.
    """

    __slots__ = ()

    #
    # Implementations can choose to define these
    # properties as using __slots__, __dict__, or
    # by overriding the @property method
    #

    variables = _abstract_readwrite_property(doc="The sos variables")
    weights = _abstract_readwrite_property(doc="The sos variables")
    level = _abstract_readwrite_property(doc="The sos level (e.g., 1,2,...)")

    #
    # Interface
    #

    def items(self):
        """Iterator over the sos variables and weights as tuples"""
        return zip(self.variables, self.weights)

    def __contains__(self, v):
        """Check if the sos contains the variable v"""
        for x in self.variables:
            if id(x) == id(v):
                return True

    def __len__(self):
        """The number of members in the set"""
        return len(self.variables)


class sos(ISOS):
    """A Special Ordered Set of type n."""

    _ctype = ISOS
    __slots__ = (
        "_parent",
        "_storage_key",
        "_active",
        "_variables",
        "_weights",
        "_level",
        "__weakref__",
    )

    def __init__(self, variables, weights=None, level=1):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._variables = tuple(variables)
        self._weights = None
        self._level = level
        if weights is None:
            self._weights = tuple(range(1, len(self._variables) + 1))
        else:
            self._weights = tuple(weights)
            for w in self._weights:
                if not is_numeric_data(w):
                    raise ValueError(
                        "Weights for Special Ordered Sets must be "
                        "expressions restricted to numeric data"
                    )

        assert len(self._variables) == len(self._weights)
        assert self._level >= 1

    #
    # Define the ISOS abstract methods
    #

    @property
    def variables(self):
        return self._variables

    @property
    def weights(self):
        return self._weights

    @property
    def level(self):
        return self._level


def sos1(variables, weights=None):
    """A Special Ordered Set of type 1.

    This is an alias for sos(..., level=1)"""
    return sos(variables, weights=weights, level=1)


def sos2(variables, weights=None):
    """A Special Ordered Set of type 2.

    This is an alias for sos(..., level=2).
    """
    return sos(variables, weights=weights, level=2)


# inserts class definitions for simple _tuple, _list, and
# _dict containers into this module
define_simple_containers(globals(), "sos", ISOS)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests detection of fixed variables."""

import pyomo.common.unittest as unittest
from pyomo.environ import ConcreteModel, TransformationFactory, Var, value


class TestDetectFixedVars(unittest.TestCase):
    """Tests detection of fixed variables."""

    def test_fixed_var_propagate(self):
        """Test for detecting de-facto fixed variables and fixing them."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v1.setub(2)
        m.v1.setlb(2)

        TransformationFactory('contrib.detect_fixed_vars').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertFalse(m.v2.fixed)

    def test_fixed_var_revert(self):
        """Test for reversion of fixed variables."""
        m = ConcreteModel()
        m.v1 = Var(initialize=1)
        m.v2 = Var(initialize=2)
        m.v1.setub(2)
        m.v1.setlb(2)

        xfrm = TransformationFactory('contrib.detect_fixed_vars')
        xfrm.apply_to(m, tmp=True)
        self.assertTrue(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        xfrm.revert(m)
        self.assertFalse(m.v1.fixed)
        self.assertEqual(value(m.v1), 1)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

import inspect

from pyomo.common.pyomo_typing import get_overloads_for
from pyomo.environ import Block


class TestTyping(unittest.TestCase):
    def test_get_overloads_for(self):
        func_list = get_overloads_for(Block.__init__)
        self.assertEqual(len(func_list), 1)
        kwds = inspect.getfullargspec(func_list[0]).kwonlyargs
        self.assertEqual(kwds, ['rule', 'concrete', 'dense', 'name', 'doc'])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.set import Set
from pyomo.contrib.mpc.data.get_cuid import get_indexed_cuid


def _is_iterable(obj):
    if not hasattr(obj, "__iter__") and not hasattr(obj, "__getitem__"):
        # Here we know obj is not iterable.
        return False
    elif hasattr(obj, "is_indexed"):
        # Pyomo scalar components define __iter__ and __getitem__,
        # however for our purpose we would like to consider them
        # as not iterable.
        #
        # Note that sets implement is_indexed but are always iterable.
        return obj.is_indexed() or isinstance(obj, Set)
    else:
        try:
            iter(obj)
            return True
        except TypeError as err:
            if "not iterable" in str(err):
                # Hopefully this error message is not implementation
                # or version specific. Tested in CPython 3.7.8 and
                # PyPy 3.7.10.
                return False
            else:
                raise err


class _DynamicDataBase:
    """
    A base class for storing data associated with time-indexed variables.

    """

    def __init__(self, data, time_set=None, context=None):
        """
        Processes keys of the data dict.

        """
        # This is used if we ever need to process a VarData to get
        # a time-indexed CUID. We need to know what set to slice.
        self._orig_time_set = time_set

        self._data = {
            get_indexed_cuid(key, (self._orig_time_set,), context=context): val
            for key, val in data.items()
        }

    def __eq__(self, other):
        if isinstance(other, _DynamicDataBase):
            return self._data == other._data
        else:
            # Should this return False or raise TypeError?
            raise TypeError(
                "%s and %s are not comparable" % (self.__class__, other.__class__)
            )

    def get_data(self):
        """
        Return a dictionary mapping CUIDs to values

        """
        return self._data

    def get_cuid(self, key, context=None):
        """
        Get the time-indexed CUID corresponding to the provided key
        """
        return get_indexed_cuid(key, (self._orig_time_set,), context=context)

    def get_data_from_key(self, key, context=None):
        """
        Returns the value associated with the given key.

        """
        cuid = get_indexed_cuid(key, (self._orig_time_set,), context=context)
        return self._data[cuid]

    def contains_key(self, key, context=None):
        """
        Returns whether this object's dict contains the given key.

        """
        cuid = get_indexed_cuid(key, (self._orig_time_set,), context=context)
        return cuid in self._data

    def update_data(self, other, context=None):
        """
        Updates this object's data dict.

        """
        if isinstance(other, _DynamicDataBase):
            self._data.update(other.get_data())
        else:
            other = {
                get_indexed_cuid(key, (self._orig_time_set,), context=context): val
                for key, val in other.items()
            }
            self._data.update(other)

    def to_serializable(self):
        """
        Returns a json-serializable object.

        """
        # We have no idea whether the values in this object's dict are
        # json-serializable.
        raise NotImplementedError(
            "to_serializable has not been implemented by %s" % self.__class__
        )

    def extract_variables(self, variables, context=None, copy_values=False):
        """
        Return a new object that only keeps data values for the variables
        specified.

        """
        if copy_values:
            # We don't know what this object uses as values in its dict,
            # so we don't know how to copy them.
            raise NotImplementedError(
                "extract_variables with copy_values=True has not been"
                " implemented by %s" % self.__class__
            )
        data = {}
        for var in variables:
            cuid = get_indexed_cuid(var, (self._orig_time_set,), context=context)
            data[cuid] = self._data[cuid]
        MyClass = self.__class__
        # Subclasses likely have different construction signatures,
        # so this maybe shouldn't be implemented on the base class.
        return MyClass(data, time_set=self._orig_time_set)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# pyros.py: Generalized Robust Cutting-Set Algorithm for Pyomo
from datetime import datetime, timezone
import logging

from pyomo.common.config import document_kwargs_from_configdict
from pyomo.core.expr import value
from pyomo.opt import SolverFactory

from pyomo.contrib.pyros.config import pyros_config, logger_domain
from pyomo.contrib.pyros.pyros_algorithm_methods import ROSolver_iterative_solve
from pyomo.contrib.pyros.solve_data import ROSolveResults
from pyomo.contrib.pyros.util import (
    load_final_solution,
    pyrosTerminationCondition,
    validate_pyros_inputs,
    log_preprocessed_model_statistics,
    log_original_model_statistics,
    IterationLogRecord,
    setup_pyros_logger,
    time_code,
    TimingData,
    ModelData,
)

__version__ = "1.3.13"


default_pyros_solver_logger = setup_pyros_logger()


def _get_pyomo_version_info():
    """
    Get Pyomo version information.
    """
    import os
    import subprocess
    from pyomo.version import version

    pyomo_version = version
    commit_hash = "unknown"

    pyros_dir = os.path.join(*os.path.split(__file__)[:-1])
    commit_hash_command_args = [
        "git",
        "-C",
        f"{pyros_dir}",
        "rev-parse",
        "--short",
        "HEAD",
    ]
    try:
        commit_hash = (
            subprocess.check_output(
                commit_hash_command_args,
                # suppress git error if Pyomo installation
                # is not a git repo
                stderr=subprocess.DEVNULL,
            )
            .decode("ascii")
            .strip()
        )
    except subprocess.CalledProcessError:
        commit_hash = "unknown"

    return {"Pyomo version": pyomo_version, "Commit hash": commit_hash}


@SolverFactory.register(
    "pyros",
    doc="Pyomo Robust Optimization Solver (PyROS): "
    "implementation of a generalized robust cutting-set algorithm (GRCS)",
)
class PyROS(object):
    """
    Pyomo Robust Optimization Solver (PyROS): implementation of a
    generalized robust cutting-set algorithm (GRCS)
    for the solution of two-stage nonlinear programs
    under uncertainty.

    We recommend instantiating this class as follows:

    .. code::

       >>> import pyomo.environ as pyo
       >>> import pyomo.contrib.pyros as pyros
       >>> pyros_solver = pyo.SolverFactory("pyros")

    """

    CONFIG = pyros_config()
    _LOG_LINE_LENGTH = 78
    _DEFAULT_CONFIG_USER_OPTIONS = [
        "first_stage_variables",
        "second_stage_variables",
        "uncertain_params",
        "uncertainty_set",
        "local_solver",
        "global_solver",
    ]

    def available(self, exception_flag=True):
        """Check if solver is available."""
        return True

    def version(self):
        """Return a 3-tuple describing the solver version."""
        return __version__

    def license_is_valid(self):
        '''License for using PyROS'''
        return True

    # The Pyomo solver API expects that solvers support the context
    # manager API
    def __enter__(self):
        return self

    def __exit__(self, et, ev, tb):
        pass

    def _log_intro(self, logger, **log_kwargs):
        """
        Log PyROS solver introductory messages.

        Parameters
        ----------
        logger : logging.Logger
            Logger through which to emit messages.
        **log_kwargs : dict, optional
            Keyword arguments to ``logger.log()`` callable.
            Should not include `msg`.
        """
        logger.log(msg="=" * self._LOG_LINE_LENGTH, **log_kwargs)
        logger.log(
            msg=f"PyROS: The Pyomo Robust Optimization Solver, v{self.version()}.",
            **log_kwargs,
        )

        # git_info_str = ", ".join(
        #     f"{field}: {val}" for field, val in _get_pyomo_git_info().items()
        # )
        version_info = _get_pyomo_version_info()
        version_info_str = ' ' * len("PyROS: ") + ("\n" + ' ' * len("PyROS: ")).join(
            f"{key}: {val}" for key, val in version_info.items()
        )
        logger.log(msg=version_info_str, **log_kwargs)
        logger.log(
            msg=(
                f"{' ' * len('PyROS:')} "
                "Invoked at UTC "
                f"{datetime.now(timezone.utc).isoformat()}"
            ),
            **log_kwargs,
        )
        logger.log(msg="", **log_kwargs)
        logger.log(
            msg=("Developed by: Natalie M. Isenberg (1), Jason A. F. Sherman (1),"),
            **log_kwargs,
        )
        logger.log(
            msg=(
                f"{' ' * len('Developed by:')} "
                "John D. Siirola (2), Chrysanthos E. Gounaris (1)"
            ),
            **log_kwargs,
        )
        logger.log(
            msg=(
                "(1) Carnegie Mellon University, " "Department of Chemical Engineering"
            ),
            **log_kwargs,
        )
        logger.log(
            msg="(2) Sandia National Laboratories, Center for Computing Research",
            **log_kwargs,
        )
        logger.log(msg="", **log_kwargs)
        logger.log(
            msg=(
                "The developers gratefully acknowledge support "
                "from the U.S. Department"
            ),
            **log_kwargs,
        )
        logger.log(
            msg=(
                "of Energy's "
                "Institute for the Design of Advanced Energy Systems (IDAES)."
            ),
            **log_kwargs,
        )
        logger.log(msg="=" * self._LOG_LINE_LENGTH, **log_kwargs)

    def _log_disclaimer(self, logger, **log_kwargs):
        """
        Log PyROS solver disclaimer messages.

        Parameters
        ----------
        logger : logging.Logger
            Logger through which to emit messages.
        **log_kwargs : dict, optional
            Keyword arguments to ``logger.log()`` callable.
            Should not include `msg`.
        """
        disclaimer_header = " DISCLAIMER ".center(self._LOG_LINE_LENGTH, "=")

        logger.log(msg=disclaimer_header, **log_kwargs)
        logger.log(msg="PyROS is currently under active development. ", **log_kwargs)
        logger.log(
            msg=(
                "Please provide feedback and/or report any issues by creating "
                "a ticket at"
            ),
            **log_kwargs,
        )
        logger.log(msg="https://github.com/Pyomo/pyomo/issues/new/choose", **log_kwargs)
        logger.log(msg="=" * self._LOG_LINE_LENGTH, **log_kwargs)

    def _log_config_user_values(
        self, logger, config, exclude_options=None, **log_kwargs
    ):
        """
        Log explicitly set PyROS solver options.

        If there are no such options, or all such options
        are to be excluded from consideration, then nothing is logged.

        Parameters
        ----------
        logger : logging.Logger
            Logger for the solver options.
        config : ConfigDict
            PyROS solver options.
        exclude_options : None or iterable of str, optional
            Options (keys of the ConfigDict) to exclude from
            logging. If `None` passed, then the names of the
            required arguments to ``self.solve()`` are skipped.
        **log_kwargs : dict, optional
            Keyword arguments to each statement of ``logger.log()``.
        """
        if exclude_options is None:
            exclude_options = set(self._DEFAULT_CONFIG_USER_OPTIONS)
        else:
            exclude_options = set(exclude_options)

        user_values = list(
            filter(lambda val: val.name() not in exclude_options, config.user_values())
        )
        if user_values:
            logger.log(msg="User-provided solver options:", **log_kwargs)
            for val in user_values:
                val_name, val_value = val.name(), val.value()
                logger.log(msg=f" {val_name}={val_value!r}", **log_kwargs)
            logger.log(msg="-" * self._LOG_LINE_LENGTH, **log_kwargs)

    def _log_config(self, logger, config, exclude_options=None, **log_kwargs):
        """
        Log PyROS solver options.

        Parameters
        ----------
        logger : logging.Logger
            Logger for the solver options.
        config : ConfigDict
            PyROS solver options.
        exclude_options : None or iterable of str, optional
            Options (keys of the ConfigDict) to exclude from
            logging. If `None` passed, then the names of the
            required arguments to ``self.solve()`` are skipped.
        **log_kwargs : dict, optional
            Keyword arguments to each statement of ``logger.log()``.
        """
        if exclude_options is None:
            exclude_options = set(self._DEFAULT_CONFIG_USER_OPTIONS)
        else:
            exclude_options = set(exclude_options)

        logger.log(msg="Full solver options:", **log_kwargs)
        for key, val in config.items():
            if key not in exclude_options:
                logger.log(msg=f" {key}={val!r}", **log_kwargs)
        logger.log(msg="-" * self._LOG_LINE_LENGTH, **log_kwargs)

    def _resolve_and_validate_pyros_args(self, model, **kwds):
        """
        Resolve and validate arguments to ``self.solve()``.

        Parameters
        ----------
        model : ConcreteModel
            Deterministic model object passed to ``self.solve()``.
        **kwds : dict
            All other arguments to ``self.solve()``.

        Returns
        -------
        config : ConfigDict
            Standardized arguments.
        user_var_partitioning : util.VarPartitioning
            User-based partitioning of the in-scope model variables.

        Note
        ----
        This method can be broken down into three steps:

        1. Cast arguments to ConfigDict. Argument-wise
           validation is performed automatically.
           Note that arguments specified directly take
           precedence over arguments specified indirectly
           through direct argument 'options'.
        2. Inter-argument validation.
        """
        # prioritize entries of kwds over entries of kwds['options']
        kwargs = {**kwds.pop("options", {}), **kwds}
        config = self.CONFIG(value=kwargs)
        user_var_partitioning = validate_pyros_inputs(model, config)
        return config, user_var_partitioning

    @document_kwargs_from_configdict(
        config=CONFIG,
        section="Keyword Arguments",
        indent_spacing=4,
        width=72,
        visibility=0,
    )
    def solve(
        self,
        model,
        first_stage_variables,
        second_stage_variables,
        uncertain_params,
        uncertainty_set,
        local_solver,
        global_solver,
        **kwds,
    ):
        """Solve a model.

        Parameters
        ----------
        model: ConcreteModel
            The deterministic model.
        first_stage_variables: VarData, Var, or iterable of VarData/Var
            First-stage model variables (or design variables).
        second_stage_variables: VarData, Var, or iterable of VarData/Var
            Second-stage model variables (or control variables).
        uncertain_params: (iterable of) Param, Var, ParamData, or VarData
            Uncertain model parameters.
            Of every constituent `Param` object,
            the `mutable` attribute must be set to True.
            All constituent `Var`/`VarData` objects should be
            fixed.
        uncertainty_set: UncertaintySet
            Uncertainty set against which the solution(s) returned
            will be confirmed to be robust.
        local_solver: str or solver type
            Subordinate local NLP solver.
            If a `str` is passed, then the `str` is cast to
            ``SolverFactory(local_solver)``.
        global_solver: str or solver type
            Subordinate global NLP solver.
            If a `str` is passed, then the `str` is cast to
            ``SolverFactory(global_solver)``.

        Returns
        -------
        return_soln : ROSolveResults
            Summary of PyROS termination outcome.

        """
        # use this to determine whether user provided
        # nominal uncertain parameter values
        nominal_param_vals_in_kwds = (
            "nominal_uncertain_param_vals" in kwds
            or "nominal_uncertain_param_vals" in kwds.get("options", {})
        )

        model_data = ModelData(original_model=model, timing=TimingData(), config=None)
        with time_code(
            timing_data_obj=model_data.timing,
            code_block_name="main",
            is_main_timer=True,
        ):
            kwds.update(
                dict(
                    first_stage_variables=first_stage_variables,
                    second_stage_variables=second_stage_variables,
                    uncertain_params=uncertain_params,
                    uncertainty_set=uncertainty_set,
                    local_solver=local_solver,
                    global_solver=global_solver,
                )
            )

            # we want to log the intro and disclaimer in
            # advance of assembling the config.
            # this helps clarify to the user that any
            # messages logged during assembly of the config
            # were, in fact, logged after PyROS was initiated
            progress_logger = logger_domain(
                kwds.get(
                    "progress_logger",
                    kwds.get("options", dict()).get(
                        "progress_logger", default_pyros_solver_logger
                    ),
                )
            )
            self._log_intro(logger=progress_logger, level=logging.INFO)
            self._log_disclaimer(logger=progress_logger, level=logging.INFO)

            config, user_var_partitioning = self._resolve_and_validate_pyros_args(
                model, **kwds
            )
            self._log_config_user_values(
                logger=config.progress_logger,
                config=config,
                exclude_options=(
                    self._DEFAULT_CONFIG_USER_OPTIONS
                    + ["nominal_uncertain_param_vals"]
                    * (not nominal_param_vals_in_kwds)
                ),
                level=logging.INFO,
            )
            self._log_config(
                logger=config.progress_logger,
                config=config,
                exclude_options=None,
                level=logging.DEBUG,
            )
            model_data.config = config

            log_original_model_statistics(model_data, user_var_partitioning)
            IterationLogRecord.log_header_rule(config.progress_logger.info)
            config.progress_logger.info("Preprocessing...")
            model_data.timing.start_timer("main.preprocessing")
            robust_infeasible = model_data.preprocess(user_var_partitioning)
            model_data.timing.stop_timer("main.preprocessing")
            preprocessing_time = model_data.timing.get_total_time("main.preprocessing")
            config.progress_logger.info(
                f"Done preprocessing; required wall time of "
                f"{preprocessing_time:.3f}s."
            )

            IterationLogRecord.log_header_rule(config.progress_logger.debug)
            log_preprocessed_model_statistics(model_data)

            # === Solve and load solution into model
            return_soln = ROSolveResults()
            if not robust_infeasible:
                pyros_soln = ROSolver_iterative_solve(model_data)
                IterationLogRecord.log_header_rule(config.progress_logger.info)

                termination_acceptable = pyros_soln.pyros_termination_condition in {
                    pyrosTerminationCondition.robust_optimal,
                    pyrosTerminationCondition.robust_feasible,
                }
                if termination_acceptable:
                    load_final_solution(
                        model_data=model_data,
                        master_soln=pyros_soln.master_results,
                        original_user_var_partitioning=user_var_partitioning,
                    )

                # get the most recent master objective, if available
                return_soln.final_objective_value = None
                master_epigraph_obj_value = value(
                    pyros_soln.master_results.master_model.epigraph_obj, exception=False
                )
                if master_epigraph_obj_value is not None:
                    # account for sense of the original model objective
                    # when reporting the final PyROS (master) objective,
                    # since maximization objective is changed to
                    # minimization objective during preprocessing
                    return_soln.final_objective_value = (
                        model_data.active_obj_original_sense * master_epigraph_obj_value
                    )

                return_soln.pyros_termination_condition = (
                    pyros_soln.pyros_termination_condition
                )
                return_soln.iterations = pyros_soln.iterations
            else:
                return_soln.final_objective_value = None
                return_soln.pyros_termination_condition = (
                    pyrosTerminationCondition.robust_infeasible
                )
                return_soln.iterations = 0

        return_soln.config = config
        return_soln.time = model_data.timing.get_total_time("main")

        # log termination-related messages
        config.progress_logger.info(return_soln.pyros_termination_condition.message)
        config.progress_logger.info("-" * self._LOG_LINE_LENGTH)
        config.progress_logger.debug(f"Timing breakdown:\n\n{model_data.timing}")
        config.progress_logger.debug("-" * self._LOG_LINE_LENGTH)
        config.progress_logger.info(return_soln)
        config.progress_logger.info("-" * self._LOG_LINE_LENGTH)
        config.progress_logger.info("All done. Exiting PyROS.")
        config.progress_logger.info("=" * self._LOG_LINE_LENGTH)

        return return_soln



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .gdp_to_mip_transformation import GDP_to_MIP_Transformation
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.core.base import TransformationFactory
from pyomo.core.util import target_list
from pyomo.gdp import Disjunction
from weakref import ref as weakref_ref
import logging

logger = logging.getLogger(__name__)


@TransformationFactory.register(
    'gdp.binary_multiplication',
    doc="Reformulate the GDP as an MINLP by multiplying f(x) <= 0 by y to get "
    "f(x) * y <= 0 where y is the binary corresponding to the Boolean indicator "
    "var of the Disjunct containing f(x) <= 0.",
)
class GDPBinaryMultiplicationTransformation(GDP_to_MIP_Transformation):
    CONFIG = ConfigDict("gdp.binary_multiplication")
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets that will be transformed",
            doc="""

        This specifies the list of components to transform. If None (default), the
        entire model is transformed. Note that if the transformation is done out
        of place, the list of targets should be attached to the model before it
        is cloned, and the list will specify the targets on the cloned
        instance.""",
        ),
    )

    transformation_name = 'binary_multiplication'

    def __init__(self):
        super().__init__(logger)

    def _apply_to(self, instance, **kwds):
        try:
            self._apply_to_impl(instance, **kwds)
        finally:
            self._restore_state()

    def _apply_to_impl(self, instance, **kwds):
        self._process_arguments(instance, **kwds)

        # filter out inactive targets and handle case where targets aren't
        # specified.
        targets = self._filter_targets(instance)
        # transform logical constraints based on targets
        self._transform_logical_constraints(instance, targets)
        # we need to preprocess targets to make sure that if there are any
        # disjunctions in targets that their disjuncts appear before them in
        # the list.
        gdp_tree = self._get_gdp_tree_from_targets(instance, targets)
        preprocessed_targets = gdp_tree.reverse_topological_sort()

        for t in preprocessed_targets:
            if t.ctype is Disjunction:
                self._transform_disjunctionData(
                    t,
                    t.index(),
                    parent_disjunct=gdp_tree.parent(t),
                    root_disjunct=gdp_tree.root_disjunct(t),
                )

    def _transform_disjunctionData(
        self, obj, index, parent_disjunct=None, root_disjunct=None
    ):
        transBlock, xorConstraint = self._setup_transform_disjunctionData(
            obj, root_disjunct
        )

        # add or (or xor) constraint
        or_expr = 0
        for disjunct in obj.disjuncts:
            or_expr += disjunct.binary_indicator_var
            self._transform_disjunct(disjunct, transBlock)

        if obj.xor:
            xorConstraint[index] = or_expr == 1
        else:
            xorConstraint[index] = or_expr >= 1
        # Mark the DisjunctionData as transformed by mapping it to its XOR
        # constraint.
        obj._algebraic_constraint = weakref_ref(xorConstraint[index])

        # and deactivate for the writers
        obj.deactivate()

    def _transform_disjunct(self, obj, transBlock):
        # We're not using the preprocessed list here, so this could be
        # inactive. We've already done the error checking in preprocessing, so
        # we just skip it here.
        if not obj.active:
            return

        relaxationBlock = self._get_disjunct_transformation_block(obj, transBlock)

        # Transform each component within this disjunct
        self._transform_block_components(obj, obj)

        # deactivate disjunct to keep the writers happy
        obj._deactivate_without_fixing_indicator()

    def _transform_constraint(self, obj, disjunct):
        # add constraint to the transformation block, we'll transform it there.
        transBlock = disjunct._transformation_block()
        constraint_map = transBlock.private_data('pyomo.gdp')

        disjunctionRelaxationBlock = transBlock.parent_block()

        # We will make indexes from ({obj.local_name} x obj.index_set() x ['lb',
        # 'ub']), but don't bother construct that set here, as taking Cartesian
        # products is kind of expensive (and redundant since we have the
        # original model)
        newConstraint = transBlock.transformedConstraints

        for i in sorted(obj.keys()):
            c = obj[i]
            if not c.active:
                continue

            self._add_constraint_expressions(
                c, i, disjunct.binary_indicator_var, newConstraint, constraint_map
            )

            # deactivate because we relaxed
            c.deactivate()

    def _add_constraint_expressions(
        self, c, i, indicator_var, newConstraint, constraint_map
    ):
        # Since we are both combining components from multiple blocks and using
        # local names, we need to make sure that the first index for
        # transformedConstraints is guaranteed to be unique. We just grab the
        # current length of the list here since that will be monotonically
        # increasing and hence unique. We'll append it to the
        # slightly-more-human-readable constraint name for something familiar
        # but unique. (Note that we really could do this outside of the loop
        # over the constraint indices, but I don't think it matters a lot.)
        unique = len(newConstraint)
        name = c.local_name + "_%s" % unique
        transformed = constraint_map.transformed_constraints[c]

        lb, ub = c.lower, c.upper
        if (c.equality or lb is ub) and lb is not None:
            # equality
            newConstraint.add((name, i, 'eq'), (c.body - lb) * indicator_var == 0)
            transformed.append(newConstraint[name, i, 'eq'])
            constraint_map.src_constraint[newConstraint[name, i, 'eq']] = c
        else:
            # inequality
            if lb is not None:
                newConstraint.add((name, i, 'lb'), 0 <= (c.body - lb) * indicator_var)
                transformed.append(newConstraint[name, i, 'lb'])
                constraint_map.src_constraint[newConstraint[name, i, 'lb']] = c
            if ub is not None:
                newConstraint.add((name, i, 'ub'), (c.body - ub) * indicator_var <= 0)
                transformed.append(newConstraint[name, i, 'ub'])
                constraint_map.src_constraint[newConstraint[name, i, 'ub']] = c



# -*- coding: utf-8 -*-

# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Implementation of the MindtPy solver.

22.2.10 changes:
- Add support for partitioning nonlinear-sum objective.

22.1.12 changes:
- Improve the log.

21.12.15 changes:
- Improve the online doc.

21.11.10 changes:
- Add support for solution pool of MIP solvers.

21.8.21 changes:
- Add support for gurobi_persistent solver in (Regularized) LP/NLP-based B&B algorithm.

21.5.19 changes:
- Add Feasibility Pump strategy.
- Add Regularized Outer Approximation method.
- Restructure and simplify the MindyPy code.

20.10.15 changes:
- Add Extended Cutting Plane and Global Outer Approximation strategy.
- Update online doc.

20.6.30 changes:
- Add support for different norms (L1, L2, L-infinity) of the objective function in the feasibility subproblem.
- Add support for different differentiate_mode to calculate Jacobian.

20.6.9 changes:
- Add cycling check in Outer Approximation method.
- Add support for GAMS solvers interface.
- Fix warmstart for both OA and LP/NLP method.

20.5.9 changes:
- Add single-tree implementation.
- Add support for cplex_persistent solver.
- Fix bug in OA cut expression in cut_generation.py.

24.1.11 changes:
- fix gurobi single tree termination check bug
- fix Gurobi single tree cycle handling
- fix bug in feasibility pump method
- add special handling for infeasible relaxed NLP
- update the log format of infeasible fixed NLP subproblems
- create a new copy_var_list_values function
"""

from pyomo.contrib.mindtpy import __version__
from pyomo.opt import SolverFactory
from pyomo.contrib.mindtpy.config_options import _get_MindtPy_config
from pyomo.common.config import document_kwargs_from_configdict
from pyomo.contrib.mindtpy.config_options import _supported_algorithms


@SolverFactory.register(
    'mindtpy', doc='MindtPy: Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo'
)
class MindtPySolver:
    """
    Decomposition solver for Mixed-Integer Nonlinear Programming (MINLP) problems.

    The MindtPy (Mixed-Integer Nonlinear Decomposition Toolbox in Pyomo) solver
    applies a variety of decomposition-based approaches to solve Mixed-Integer
    Nonlinear Programming (MINLP) problems.
    These approaches include:

    - Outer approximation (OA)
    - Global outer approximation (GOA)
    - Regularized outer approximation (ROA)
    - LP/NLP based branch-and-bound (LP/NLP)
    - Global LP/NLP based branch-and-bound (GLP/NLP)
    - Regularized LP/NLP based branch-and-bound (RLP/NLP)
    - Feasibility pump (FP)
    """

    CONFIG = _get_MindtPy_config()

    def available(self, exception_flag=True):
        """Check if solver is available."""
        return True

    def license_is_valid(self):
        return True

    def version(self):
        """Return a 3-tuple describing the solver version."""
        return __version__

    @document_kwargs_from_configdict(CONFIG)
    def solve(self, model, **kwds):
        """Solve the model.

        Args:
            model (Block): a Pyomo model or block to be solved

        """
        # The algorithm should have been specified as an argument to the solve
        # method. We will instantiate an ephemeral instance of the correct
        # solver and call its solve method.
        options = kwds.pop('options', {})
        config = self.CONFIG(options, preserve_implicit=True)
        # Don't complain about extra things, they aren't for us. We just need to
        # get the algorithm and then our job is done.
        config.set_value(kwds, skip_implicit=True)

        return SolverFactory(_supported_algorithms[config.strategy][0]).solve(
            model, **kwds
        )

    #
    # Support 'with' statements.
    #
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import pyomo.common.unittest as unittest
from unittest import mock
from pyomo.environ import (
    ConcreteModel,
    Var,
    Objective,
    Constraint,
    NonNegativeIntegers,
    NonNegativeReals,
    Reals,
    Integers,
    maximize,
    minimize,
    Suffix,
)
from pyomo.opt.results import SolverStatus, TerminationCondition, ProblemSense
from pyomo.opt import SolverFactory
import warnings

CFGFILE = os.environ.get("SAS_CFG_FILE_PATH", None)

CAS_OPTIONS = {
    "hostname": os.environ.get("CASHOST", None),
    "port": os.environ.get("CASPORT", None),
    "authinfo": os.environ.get("CASAUTHINFO", None),
}


try:
    sas94_available = SolverFactory('_sas94').available()
except:
    sas94_available = False


class SASTestAbc:
    solver_io = "_sas94"
    session_options = {}
    cfgfile = CFGFILE

    @classmethod
    def setUpClass(cls):
        cls.opt_sas = SolverFactory(
            "sas", solver_io=cls.solver_io, cfgfile=cls.cfgfile, **cls.session_options
        )

    @classmethod
    def tearDownClass(cls):
        del cls.opt_sas

    def setObj(self):
        X = self.instance.X
        self.instance.Obj = Objective(
            expr=2 * X[1] - 3 * X[2] - 4 * X[3], sense=minimize
        )

    def setX(self):
        self.instance.X = Var([1, 2, 3], within=NonNegativeReals)

    def setUp(self):
        # Disable resource warnings
        warnings.filterwarnings("ignore", category=ResourceWarning)
        instance = self.instance = ConcreteModel()
        self.setX()
        X = instance.X
        instance.R1 = Constraint(expr=-2 * X[2] - 3 * X[3] >= -5)
        instance.R2 = Constraint(expr=X[1] + X[2] + 2 * X[3] <= 4)
        instance.R3 = Constraint(expr=X[1] + 2 * X[2] + 3 * X[3] <= 7)
        self.setObj()

        # Declare suffixes for solution information
        instance.status = Suffix(direction=Suffix.IMPORT)
        instance.slack = Suffix(direction=Suffix.IMPORT)
        instance.rc = Suffix(direction=Suffix.IMPORT)
        instance.dual = Suffix(direction=Suffix.IMPORT)

    def tearDown(self):
        del self.instance

    def run_solver(self, **kwargs):
        opt_sas = self.opt_sas
        instance = self.instance

        # Call the solver
        self.results = opt_sas.solve(instance, **kwargs)


class SASTestLP(SASTestAbc):
    def checkSolution(self):
        instance = self.instance
        results = self.results
        # Get the objective sense, we use the same code for minimization and maximization tests
        sense = instance.Obj.sense

        # Check status
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )

        # Check objective value
        self.assertAlmostEqual(instance.Obj(), sense * -7.5)

        # Check primal solution values
        self.assertAlmostEqual(instance.X[1].value, 0.0)
        self.assertAlmostEqual(instance.X[2].value, 2.5)
        self.assertAlmostEqual(instance.X[3].value, 0.0)

        # Check reduced cost
        self.assertAlmostEqual(instance.rc[instance.X[1]], sense * 2.0)
        self.assertAlmostEqual(instance.rc[instance.X[2]], sense * 0.0)
        self.assertAlmostEqual(instance.rc[instance.X[3]], sense * 0.5)

        # Check slack
        self.assertAlmostEqual(instance.slack[instance.R1], -5.0)
        self.assertAlmostEqual(instance.slack[instance.R2], 2.5)
        self.assertAlmostEqual(instance.slack[instance.R3], 5.0)

        # Check dual solution
        self.assertAlmostEqual(instance.dual[instance.R1], sense * 1.5)
        self.assertAlmostEqual(instance.dual[instance.R2], sense * 0.0)
        self.assertAlmostEqual(instance.dual[instance.R3], sense * 0.0)

        # Check basis status
        self.assertEqual(instance.status[instance.X[1]], "L")
        self.assertEqual(instance.status[instance.X[2]], "B")
        self.assertEqual(instance.status[instance.X[3]], "L")
        self.assertEqual(instance.status[instance.R1], "U")
        self.assertEqual(instance.status[instance.R2], "B")
        self.assertEqual(instance.status[instance.R3], "B")

    def test_solver_default(self):
        self.run_solver()
        self.checkSolution()

    def test_solver_tee(self):
        self.run_solver(tee=True)
        self.checkSolution()

    def test_solver_primal(self):
        self.run_solver(options={"algorithm": "ps"})
        self.assertIn("NOTE: The Primal Simplex algorithm is used.", self.opt_sas._log)
        self.checkSolution()

    def test_solver_ipm(self):
        self.run_solver(options={"algorithm": "ip"})
        self.assertIn("NOTE: The Interior Point algorithm is used.", self.opt_sas._log)
        self.checkSolution()

    def test_solver_intoption(self):
        self.run_solver(options={"maxiter": 20})
        self.checkSolution()

    def test_solver_invalidoption(self):
        with self.assertRaisesRegex(ValueError, "syntax error"):
            self.run_solver(options={"foo": "bar"})

    def test_solver_max(self):
        X = self.instance.X
        self.instance.Obj.set_value(expr=-2 * X[1] + 3 * X[2] + 4 * X[3])
        self.instance.Obj.sense = maximize
        self.run_solver()
        self.checkSolution()
        self.assertEqual(self.results.problem.sense, ProblemSense.maximize)

    def test_solver_infeasible(self):
        instance = self.instance
        X = instance.X
        instance.R4 = Constraint(expr=-2 * X[2] - 3 * X[3] <= -6)
        self.run_solver()
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertEqual(results.solver.message, "The problem is infeasible.")

    def test_solver_infeasible_or_unbounded(self):
        self.instance.X.domain = Reals
        self.run_solver()
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertIn(
            results.solver.termination_condition,
            [
                TerminationCondition.infeasibleOrUnbounded,
                TerminationCondition.unbounded,
            ],
        )
        self.assertIn(
            results.solver.message,
            ["The problem is infeasible or unbounded.", "The problem is unbounded."],
        )

    def test_solver_unbounded(self):
        self.instance.X.domain = Reals
        self.run_solver(options={"presolver": "none", "algorithm": "primal"})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.unbounded
        )
        self.assertEqual(results.solver.message, "The problem is unbounded.")

    def checkSolutionDecomp(self):
        instance = self.instance
        results = self.results
        # Get the objective sense, we use the same code for minimization and maximization tests
        sense = instance.Obj.sense

        # Check status
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )

        # Check objective value
        self.assertAlmostEqual(instance.Obj(), sense * -7.5)

        # Check primal solution values
        self.assertAlmostEqual(instance.X[1].value, 0.0)
        self.assertAlmostEqual(instance.X[2].value, 2.5)
        self.assertAlmostEqual(instance.X[3].value, 0.0)

        # Check reduced cost
        self.assertAlmostEqual(instance.rc[instance.X[1]], sense * 2.0)
        self.assertAlmostEqual(instance.rc[instance.X[2]], sense * 0.0)
        self.assertAlmostEqual(instance.rc[instance.X[3]], sense * 0.5)

        # Check slack
        self.assertAlmostEqual(instance.slack[instance.R1], -5.0)
        self.assertAlmostEqual(instance.slack[instance.R2], 2.5)
        self.assertAlmostEqual(instance.slack[instance.R3], 5.0)

        # Check dual solution
        self.assertAlmostEqual(instance.dual[instance.R1], sense * 1.5)
        self.assertAlmostEqual(instance.dual[instance.R2], sense * 0.0)
        self.assertAlmostEqual(instance.dual[instance.R3], sense * 0.0)

        # Don't check basis status for decomp

    def test_solver_decomp(self):
        self.run_solver(
            options={
                "decomp": {"absobjgap": 0.0},
                "decompmaster": {"algorithm": "dual"},
                "decompsubprob": {"presolver": "none"},
            }
        )
        self.assertIn(
            "NOTE: The DECOMP method value DEFAULT is applied.", self.opt_sas._log
        )
        self.checkSolutionDecomp()

    def test_solver_iis(self):
        self.run_solver(options={"iis": "true"})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.feasible
        )
        self.assertIn("NOTE: The IIS= option is enabled.", self.opt_sas._log)
        self.assertEqual(
            results.solver.message,
            "The problem is feasible. This status is displayed when the IIS=TRUE option is specified and the problem is feasible.",
        )

    def test_solver_maxiter(self):
        self.run_solver(options={"maxiter": 1})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxIterations
        )
        self.assertEqual(
            results.solver.message,
            "The maximum allowable number of iterations was reached.",
        )

    def test_solver_with_milp(self):
        self.run_solver(options={"with": "milp"})
        self.assertIn(
            "WARNING: The problem has no integer variables.", self.opt_sas._log
        )


@unittest.skipIf(not sas94_available, "The SAS94 solver interface is not available")
@unittest.pytest.mark.solver("sas")
class SASTestLP94(SASTestLP, unittest.TestCase):
    @mock.patch(
        "pyomo.solvers.plugins.solvers.SAS.SAS94.sas_version",
        return_value="9.sd45s39M4234232",
    )
    def test_solver_versionM4(self, sas):
        with self.assertRaises(NotImplementedError):
            self.run_solver()

    @mock.patch(
        "pyomo.solvers.plugins.solvers.SAS.SAS94.sas_version",
        return_value="9.34897293M5324u98",
    )
    def test_solver_versionM5(self, sas):
        self.run_solver()
        self.checkSolution()

    @mock.patch("saspy.SASsession.submit", return_value={"LOG": "", "LST": ""})
    @mock.patch("saspy.SASsession.symget", return_value="STATUS=OUT_OF_MEMORY")
    def test_solver_out_of_memory(self, submit_mock, symget_mocks):
        self.run_solver(load_solutions=False)
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.aborted)

    @mock.patch("saspy.SASsession.submit", return_value={"LOG": "", "LST": ""})
    @mock.patch("saspy.SASsession.symget", return_value="STATUS=ERROR")
    def test_solver_error(self, submit_mock, symget_mock):
        self.run_solver(load_solutions=False)
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.error)


# @unittest.skipIf(not sascas_available, "The SAS solver is not available")
@unittest.skip("Tests not yet configured for SAS Viya interface.")
@unittest.pytest.mark.solver("sas")
class SASTestLPCAS(SASTestLP, unittest.TestCase):
    solver_io = "_sascas"
    session_options = CAS_OPTIONS

    @mock.patch("pyomo.solvers.plugins.solvers.SAS.stat")
    def test_solver_large_file(self, os_stat):
        os_stat.return_value.st_size = 3 * 1024**3
        self.run_solver()
        self.checkSolution()


@unittest.pytest.mark.solver("sas")
class SASTestMILP(SASTestAbc):
    def setX(self):
        self.instance.X = Var([1, 2, 3], within=NonNegativeIntegers)

    def checkSolution(self):
        instance = self.instance
        results = self.results

        # Get the objective sense, we use the same code for minimization and maximization tests
        sense = instance.Obj.sense

        # Check status
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )

        # Check objective value
        self.assertAlmostEqual(instance.Obj(), sense * -7)

        # Check primal solution values
        self.assertAlmostEqual(instance.X[1].value, 0.0)
        self.assertAlmostEqual(instance.X[2].value, 1.0)
        self.assertAlmostEqual(instance.X[3].value, 1.0)

    def test_solver_default(self):
        self.run_solver()
        self.checkSolution()

    def test_solver_tee(self):
        self.run_solver(tee=True)
        self.checkSolution()

    def test_solver_presolve(self):
        self.run_solver(options={"presolver": "none"})
        self.assertIn(
            "NOTE: The MILP presolver value NONE is applied.", self.opt_sas._log
        )
        self.checkSolution()

    def test_solver_intoption(self):
        self.run_solver(options={"maxnodes": 20})
        self.checkSolution()

    def test_solver_invalidoption(self):
        with self.assertRaisesRegex(ValueError, "syntax error"):
            self.run_solver(options={"foo": "bar"})

    def test_solver_max(self):
        X = self.instance.X
        self.instance.Obj.set_value(expr=-2 * X[1] + 3 * X[2] + 4 * X[3])
        self.instance.Obj.sense = maximize
        self.run_solver()
        self.checkSolution()

    def test_solver_infeasible(self):
        instance = self.instance
        X = instance.X
        instance.R4 = Constraint(expr=-2 * X[2] - 3 * X[3] <= -6)
        self.run_solver()
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.infeasible
        )
        self.assertEqual(results.solver.message, "The problem is infeasible.")

    def test_solver_infeasible_or_unbounded(self):
        self.instance.X.domain = Integers
        self.run_solver()
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertIn(
            results.solver.termination_condition,
            [
                TerminationCondition.infeasibleOrUnbounded,
                TerminationCondition.unbounded,
            ],
        )
        self.assertIn(
            results.solver.message,
            ["The problem is infeasible or unbounded.", "The problem is unbounded."],
        )

    def test_solver_unbounded(self):
        self.instance.X.domain = Integers
        self.run_solver(
            options={"presolver": "none", "rootnode": {"algorithm": "primal"}}
        )
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.warning)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.unbounded
        )
        self.assertEqual(results.solver.message, "The problem is unbounded.")

    def test_solver_decomp(self):
        self.run_solver(
            options={
                "decomp": {"hybrid": "off"},
                "decompmaster": {"algorithm": "dual"},
                "decompmasterip": {"presolver": "none"},
                "decompsubprob": {"presolver": "none"},
            }
        )
        self.assertIn(
            "NOTE: The DECOMP method value DEFAULT is applied.", self.opt_sas._log
        )
        self.checkSolution()

    def test_solver_rootnode(self):
        self.run_solver(options={"rootnode": {"presolver": "automatic"}})
        self.checkSolution()

    def test_solver_maxnodes(self):
        self.run_solver(options={"maxnodes": 0})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxEvaluations
        )
        self.assertEqual(
            results.solver.message,
            "The solver reached the maximum number of nodes specified by the MAXNODES= option and found a solution.",
        )

    def test_solver_maxsols(self):
        self.run_solver(options={"maxsols": 1})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.maxEvaluations
        )
        self.assertEqual(
            results.solver.message,
            "The solver reached the maximum number of solutions specified by the MAXSOLS= option.",
        )

    def test_solver_target(self):
        self.run_solver(options={"target": -6.0})
        results = self.results
        self.assertEqual(results.solver.status, SolverStatus.ok)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )
        self.assertEqual(
            results.solver.message,
            "The solution is not worse than the target specified by the TARGET= option.",
        )

    def test_solver_primalin(self):
        X = self.instance.X
        X[1] = None
        X[2] = 3
        X[3] = 7
        self.run_solver(warmstart=True)
        self.checkSolution()
        self.assertIn(
            "NOTE: The input solution is infeasible or incomplete. Repair heuristics are applied.",
            self.opt_sas._log,
        )

    def test_solver_primalin_nosol(self):
        X = self.instance.X
        X[1] = None
        X[2] = None
        X[3] = None
        self.run_solver(warmstart=True)
        self.checkSolution()

    @mock.patch("pyomo.solvers.plugins.solvers.SAS.stat")
    def test_solver_large_file(self, os_stat):
        os_stat.return_value.st_size = 3 * 1024**3
        self.run_solver()
        self.checkSolution()

    def test_solver_with_lp(self):
        self.run_solver(options={"with": "lp"})
        self.assertIn(
            "contains integer variables; the linear relaxation will be solved.",
            self.opt_sas._log,
        )

    def test_solver_warmstart_capable(self):
        self.run_solver()
        self.assertTrue(self.opt_sas.warm_start_capable())


# @unittest.skipIf(not sas94_available, "The SAS solver is not available")
@unittest.skip("MILP94 tests disabled.")
@unittest.pytest.mark.solver("sas")
class SASTestMILP94(SASTestMILP, unittest.TestCase):
    pass


# @unittest.skipIf(not sascas_available, "The SAS solver is not available")
@unittest.skip("Tests not yet configured for SAS Viya interface.")
@unittest.pytest.mark.solver("sas")
class SASTestMILPCAS(SASTestMILP, unittest.TestCase):
    solver_io = "_sascas"
    session_options = CAS_OPTIONS


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy_available
from pyomo.common import unittest

import pyomo.contrib.alternative_solutions.tests.test_cases as tc
from pyomo.contrib.alternative_solutions import lp_enum
from pyomo.contrib.alternative_solutions import lp_enum_solnpool
from pyomo.opt import check_available_solvers

import pyomo.environ as pyo

# lp_enum_solnpool uses both 'gurobi' and 'appsi_gurobi'
gurobi_available = len(check_available_solvers('gurobi', 'appsi_gurobi')) == 2

#
# TODO: Setup detailed tests here
#


@unittest.skipUnless(gurobi_available, "Gurobi MIP solver not available")
@unittest.skipUnless(numpy_available, "NumPy not found")
class TestLPEnumSolnpool(unittest.TestCase):

    def test_here(self):
        n = tc.get_pentagonal_pyramid_mip()
        n.x.domain = pyo.Reals
        n.y.domain = pyo.Reals

        try:
            sols = lp_enum_solnpool.enumerate_linear_solutions_soln_pool(n, tee=True)
        except pyomo.common.errors.ApplicationError as e:
            sols = []

        # TODO - Confirm how solnpools deal with duplicate solutions
        if gurobi_available:
            assert len(sols) == 7
        else:
            assert len(sols) == 0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os

from pyomo.common.tempfiles import TempfileManager
import pyomo.common.unittest as unittest

import pyomo.kernel as pmo
from pyomo.core import (
    Binary,
    ConcreteModel,
    Constraint,
    Objective,
    Var,
    Integers,
    RangeSet,
    minimize,
    quicksum,
    Suffix,
)
from pyomo.opt import ProblemFormat, convert_problem, SolverFactory, BranchDirection
from pyomo.solvers.plugins.solvers.CPLEX import (
    CPLEXSHELL,
    MockCPLEX,
    _validate_file_name,
)


class _mock_cplex_128:
    def version(self):
        return (12, 8, 0)


class _mock_cplex_126:
    def version(self):
        return (12, 6, 0)


class CPLEX_utils(unittest.TestCase):
    def test_validate_file_name(self):
        _126 = _mock_cplex_126()
        _128 = _mock_cplex_128()

        # Check plain file
        fname = 'foo.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))

        # Check spaces in the file
        fname = 'foo bar.lp'
        with self.assertRaisesRegex(ValueError, "Space detected in CPLEX xxx file"):
            _validate_file_name(_126, fname, 'xxx')
        self.assertEqual('"%s"' % (fname,), _validate_file_name(_128, fname, 'xxx'))

        # check OK path separators
        fname = 'foo%sbar.lp' % (os.path.sep,)
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))

        # check BAD path separators
        bad_char = '/\\'.replace(os.path.sep, '')
        fname = 'foo%sbar.lp' % (bad_char,)
        msg = r'Unallowed character \(%s\) found in CPLEX xxx file' % (
            repr(bad_char)[1:-1],
        )
        with self.assertRaisesRegex(ValueError, msg):
            _validate_file_name(_126, fname, 'xxx')
        with self.assertRaisesRegex(ValueError, msg):
            _validate_file_name(_128, fname, 'xxx')

        # check allowable characters
        fname = 'foo$$bar.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))
        fname = 'foo_bar.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))
        fname = 'foo&bar.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))
        fname = 'foo~bar.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))
        fname = 'foo-bar.lp'
        self.assertEqual(fname, _validate_file_name(_126, fname, 'xxx'))
        self.assertEqual(fname, _validate_file_name(_128, fname, 'xxx'))

        # Check unallowable character
        bad_char = '^'
        fname = 'foo%sbar.lp' % (bad_char,)
        msg = r"Unallowed character \(\^\) found in CPLEX xxx file"
        with self.assertRaisesRegex(ValueError, msg):
            _validate_file_name(_126, fname, 'xxx')
        with self.assertRaisesRegex(ValueError, msg):
            _validate_file_name(_128, fname, 'xxx')


class CPLEXShellWritePrioritiesFile(unittest.TestCase):
    """Unit test on writing of priorities via `CPLEXSHELL._write_priorities_file()`"""

    suffix_cls = Suffix

    def setUp(self):
        TempfileManager.push()
        self.mock_model = self.get_mock_model()
        self.mock_cplex_shell = self.get_mock_cplex_shell(self.mock_model)
        self.mock_cplex_shell._priorities_file_name = TempfileManager.create_tempfile(
            suffix=".cplex.ord"
        )

    def tearDown(self):
        TempfileManager.clear_tempfiles()

    def get_mock_model(self):
        model = ConcreteModel()
        model.x = Var(within=Binary)
        model.con = Constraint(expr=model.x >= 1)
        model.obj = Objective(expr=model.x)
        return model

    def get_mock_cplex_shell(self, mock_model):
        solver = MockCPLEX()
        solver._problem_files, solver._problem_format, solver._smap_id = (
            convert_problem(
                (mock_model,),
                ProblemFormat.cpxlp,
                [ProblemFormat.cpxlp],
                has_capability=lambda x: True,
                symbolic_solver_labels=True,
            )
        )
        return solver

    def get_priorities_file_as_string(self, mock_cplex_shell):
        with open(mock_cplex_shell._priorities_file_name, "r") as ord_file:
            priorities_file = ord_file.read()
        return priorities_file

    @staticmethod
    def _set_suffix_value(suffix, variable, value):
        suffix.set_value(variable, value)

    def test_write_without_priority_suffix(self):
        with self.assertRaises(ValueError):
            CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)

    def test_write_priority_to_priorities_file(self):
        self.mock_model.priority = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        priority_val = 10
        self._set_suffix_value(
            self.mock_model.priority, self.mock_model.x, priority_val
        )

        CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)
        priorities_file = self.get_priorities_file_as_string(self.mock_cplex_shell)

        self.assertEqual(
            priorities_file,
            "* ENCODING=ISO-8859-1\n"
            "NAME             Priority Order\n"
            "  x 10\n"
            "ENDATA\n",
        )

    def test_write_priority_and_direction_to_priorities_file(self):
        self.mock_model.priority = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        priority_val = 10
        self._set_suffix_value(
            self.mock_model.priority, self.mock_model.x, priority_val
        )

        self.mock_model.direction = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        direction_val = BranchDirection.down
        self._set_suffix_value(
            self.mock_model.direction, self.mock_model.x, direction_val
        )

        CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)
        priorities_file = self.get_priorities_file_as_string(self.mock_cplex_shell)

        self.assertEqual(
            priorities_file,
            "* ENCODING=ISO-8859-1\n"
            "NAME             Priority Order\n"
            " DN x 10\n"
            "ENDATA\n",
        )

    def test_raise_due_to_invalid_priority(self):
        self.mock_model.priority = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        self._set_suffix_value(self.mock_model.priority, self.mock_model.x, -1)
        with self.assertRaises(ValueError):
            CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)

        self._set_suffix_value(self.mock_model.priority, self.mock_model.x, 1.1)
        with self.assertRaises(ValueError):
            CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)

    def test_use_default_due_to_invalid_direction(self):
        self.mock_model.priority = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        priority_val = 10
        self._set_suffix_value(
            self.mock_model.priority, self.mock_model.x, priority_val
        )

        self.mock_model.direction = self.suffix_cls(
            direction=Suffix.EXPORT, datatype=Suffix.INT
        )
        self._set_suffix_value(
            self.mock_model.direction, self.mock_model.x, "invalid_branching_direction"
        )

        CPLEXSHELL._write_priorities_file(self.mock_cplex_shell, self.mock_model)
        priorities_file = self.get_priorities_file_as_string(self.mock_cplex_shell)

        self.assertEqual(
            priorities_file,
            "* ENCODING=ISO-8859-1\n"
            "NAME             Priority Order\n"
            "  x 10\n"
            "ENDATA\n",
        )


class CPLEXShellWritePrioritiesFileKernel(CPLEXShellWritePrioritiesFile):
    suffix_cls = pmo.suffix

    @staticmethod
    def _set_suffix_value(suffix, variable, value):
        suffix[variable] = value

    def get_mock_model(self):
        model = pmo.block()
        model.x = pmo.variable(domain=Binary)
        model.con = pmo.constraint(expr=model.x >= 1)
        model.obj = pmo.objective(expr=model.x)
        return model


class CPLEXShellSolvePrioritiesFile(unittest.TestCase):
    """Integration test on the end-to-end application of priorities via the `Suffix` through a `solve()`"""

    def get_mock_model_with_priorities(self):
        m = ConcreteModel()
        m.x = Var(domain=Integers)
        m.s = RangeSet(0, 9)
        m.y = Var(m.s, domain=Integers)
        m.o = Objective(expr=m.x + sum(m.y), sense=minimize)
        m.c = Constraint(expr=m.x >= 1)
        m.c2 = Constraint(expr=quicksum(m.y[i] for i in m.s) >= 10)

        m.priority = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        m.direction = Suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)

        m.priority.set_value(m.x, 1)

        # Ensure tests work for both options of `expand`
        m.priority.set_value(m.y, 2, expand=False)
        m.direction.set_value(m.y, BranchDirection.down, expand=True)

        m.direction.set_value(m.y[9], BranchDirection.up)
        return m

    def test_use_variable_priorities(self):
        model = self.get_mock_model_with_priorities()
        with SolverFactory("_mock_cplex") as opt:
            opt._presolve(
                model, priorities=True, keepfiles=True, symbolic_solver_labels=True
            )

            with open(opt._priorities_file_name, "r") as ord_file:
                priorities_file = ord_file.read()

        self.assertEqual(
            priorities_file,
            (
                "* ENCODING=ISO-8859-1\n"
                "NAME             Priority Order\n"
                "  x 1\n"
                " DN y(0) 2\n"
                " DN y(1) 2\n"
                " DN y(2) 2\n"
                " DN y(3) 2\n"
                " DN y(4) 2\n"
                " DN y(5) 2\n"
                " DN y(6) 2\n"
                " DN y(7) 2\n"
                " DN y(8) 2\n"
                " UP y(9) 2\n"
                "ENDATA\n"
            ),
        )
        self.assertIn("read %s\n" % (opt._priorities_file_name,), opt._command.script)

    def test_ignore_variable_priorities(self):
        model = self.get_mock_model_with_priorities()
        with SolverFactory("_mock_cplex") as opt:
            opt._presolve(model, priorities=False, keepfiles=True)

            self.assertIsNone(opt._priorities_file_name)
            self.assertNotIn(".ord", opt._command.script)

    def test_can_use_manual_priorities_file_with_lp_solve(self):
        """Test that we can pass an LP file (not a pyomo model) along with a priorities file to `.solve()`"""
        model = self.get_mock_model_with_priorities()

        with SolverFactory("_mock_cplex") as pre_opt:
            pre_opt._presolve(model, priorities=True, keepfiles=True)
            lp_file = pre_opt._problem_files[0]
            priorities_file_name = pre_opt._priorities_file_name

            with open(priorities_file_name, "r") as ord_file:
                provided_priorities_file = ord_file.read()

        with SolverFactory("_mock_cplex") as opt:
            opt._presolve(
                lp_file,
                priorities=True,
                priorities_file=priorities_file_name,
                keepfiles=True,
            )

            self.assertIn(".ord", opt._command.script)

            with open(opt._priorities_file_name, "r") as ord_file:
                priorities_file = ord_file.read()

        self.assertEqual(priorities_file, provided_priorities_file)


class CPLEXShellSolvePrioritiesFileKernel(CPLEXShellSolvePrioritiesFile):
    def get_mock_model_with_priorities(self):
        m = pmo.block()
        m.x = pmo.variable(domain=Integers)
        m.s = range(10)

        m.y = pmo.variable_list(pmo.variable(domain=Integers) for _ in m.s)

        m.o = pmo.objective(expr=m.x + sum(m.y), sense=minimize)
        m.c = pmo.constraint(expr=m.x >= 1)
        m.c2 = pmo.constraint(expr=quicksum(m.y[i] for i in m.s) >= 10)

        m.priority = pmo.suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)
        m.direction = pmo.suffix(direction=Suffix.EXPORT, datatype=Suffix.INT)

        m.priority[m.x] = 1
        m.priority[m.y] = 2
        m.direction[m.y] = BranchDirection.down
        m.direction[m.y[-1]] = BranchDirection.up
        return m


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import ctypes
import logging
import os
import platform
import shutil
import stat
import sys
import tempfile
import subprocess

from io import StringIO

import pyomo.common.unittest as unittest

import pyomo.common.envvar as envvar
from pyomo.common.log import LoggingIntercept
from pyomo.common.fileutils import (
    this_file,
    this_file_dir,
    find_file,
    find_library,
    find_executable,
    PathManager,
    _system,
    _path,
    _exeExt,
    _libExt,
    ExecutableData,
    import_file,
    to_legal_filename,
)
from pyomo.common.download import FileDownloader

try:
    samefile = os.path.samefile
except AttributeError:
    # os.path.samefile is not available in Python 2.7 under Windows.
    # Mock up a dummy function for that platform.
    def samefile(a, b):
        return True


_this_file = this_file()
_this_file_dir = this_file_dir()


class TestFileUtils(unittest.TestCase):
    def setUp(self):
        self.tmpdir = None
        self.basedir = os.path.abspath(os.path.curdir)
        self.config = envvar.PYOMO_CONFIG_DIR
        self.ld_library_path = os.environ.get('LD_LIBRARY_PATH', None)
        self.path = os.environ.get('PATH', None)

    def tearDown(self):
        envvar.PYOMO_CONFIG_DIR = self.config
        os.chdir(self.basedir)
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)
        if self.ld_library_path is None:
            os.environ.pop('LD_LIBRARY_PATH', None)
        else:
            os.environ['LD_LIBRARY_PATH'] = self.ld_library_path
        if self.path is None:
            os.environ.pop('PATH', None)
        else:
            os.environ['PATH'] = self.path

    def _make_exec(self, fname):
        open(fname, 'w').close()
        mode = os.stat(fname).st_mode
        os.chmod(fname, mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    def _check_file(self, found, ref):
        #
        # On OSX, the identified path is (sometimes) prepended with
        # '/private/'. These two paths are in fact equivalent on OSX,
        # but we will explicitly verify that fact.  We check the string
        # first so that we can generate a more informative error in the
        # case of "gross" failure.
        #
        self.assertTrue(found.endswith(ref), "%s does not end with %s" % (found, ref))
        self.assertTrue(samefile(ref, found))

    def test_this_file(self):
        self.assertEqual(_this_file, __file__.replace('.pyc', '.py'))
        # Note that in some versions of PyPy, this can return <module>
        # instead of the normal <string>
        self.assertIn(
            subprocess.run(
                [
                    sys.executable,
                    '-c',
                    'from pyomo.common.fileutils import this_file;'
                    'print(this_file())',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            ).stdout.strip(),
            ['<string>', '<module>'],
        )
        self.assertEqual(
            subprocess.run(
                [sys.executable],
                input='from pyomo.common.fileutils import this_file;'
                'print(this_file())',
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                universal_newlines=True,
            ).stdout.strip(),
            '<stdin>',
        )

    def test_this_file_dir(self):
        expected_path = os.path.join('pyomo', 'common', 'tests')
        self.assertTrue(_this_file_dir.endswith(expected_path))

    def test_import_file(self):
        import_ex = import_file(os.path.join(_this_file_dir, 'import_ex.py'))
        self.assertIn("pyomo.common.tests.import_ex", sys.modules)
        self.assertEqual(import_ex.b, 2)

    def test_import_file_no_extension(self):
        with self.assertRaises(FileNotFoundError) as context:
            import_file(os.path.join(_this_file_dir, 'import_ex'))
        self.assertTrue('File does not exist' in str(context.exception))

    def test_system(self):
        self.assertTrue(platform.system().lower().startswith(_system()))
        self.assertNotIn('.', _system())
        self.assertNotIn('-', _system())
        self.assertNotIn('_', _system())

    def test_path(self):
        orig_path = os.environ.get('PATH', None)
        if orig_path:
            self.assertEqual(os.pathsep.join(_path()), os.environ['PATH'])
        os.environ.pop('PATH', None)
        self.assertEqual(os.pathsep.join(_path()), os.defpath)
        # NOTE: PATH restored by tearDown()

    def test_findfile(self):
        self.tmpdir = os.path.abspath(tempfile.mkdtemp())
        subdir_name = 'aaa'
        subdir = os.path.join(self.tmpdir, subdir_name)
        os.mkdir(subdir)
        # CWD restored in tearDown
        os.chdir(self.tmpdir)

        fname = 'foo.py'
        self.assertEqual(None, find_file(fname))

        open(os.path.join(self.tmpdir, fname), 'w').close()
        open(os.path.join(subdir, fname), 'w').close()
        open(os.path.join(subdir, 'aaa'), 'w').close()

        # we can find files in the CWD
        self._check_file(find_file(fname), os.path.join(self.tmpdir, fname))

        # unless we don't look in the cwd
        self.assertIsNone(find_file(fname, cwd=False))

        # cwd overrides pathlist
        self._check_file(
            find_file(fname, pathlist=[subdir]), os.path.join(self.tmpdir, fname)
        )

        self._check_file(
            find_file(fname, pathlist=[subdir], cwd=False), os.path.join(subdir, fname)
        )

        # ...unless the CWD match fails the MODE check
        #  (except on Windows, where all files have X_OK)
        found = find_file(fname, pathlist=[subdir], mode=os.X_OK)
        if _system() in ('windows', 'cygwin'):
            self._check_file(found, os.path.join(self.tmpdir, fname))
        else:
            self.assertIsNone(found)

        self._make_exec(os.path.join(subdir, fname))
        found = find_file(fname, pathlist=[subdir], mode=os.X_OK)
        if _system() in ('windows', 'cygwin'):
            ref = os.path.join(self.tmpdir, fname)
        else:
            ref = os.path.join(subdir, fname)
        self._check_file(found, ref)

        # pathlist may also be a string
        self._check_file(
            find_file(fname, pathlist=os.pathsep + subdir + os.pathsep, cwd=False),
            os.path.join(subdir, fname),
        )

        # implicit extensions work (even if they are not necessary)
        self._check_file(find_file(fname, ext='.py'), os.path.join(self.tmpdir, fname))
        self._check_file(
            find_file(fname, ext=['.py']), os.path.join(self.tmpdir, fname)
        )

        # implicit extensions work (when they are necessary)
        self._check_file(
            find_file(fname[:-3], ext='.py'), os.path.join(self.tmpdir, fname)
        )

        self._check_file(
            find_file(fname[:-3], ext=['.py']), os.path.join(self.tmpdir, fname)
        )

        # only files are found
        self._check_file(
            find_file(subdir_name, pathlist=[self.tmpdir, subdir], cwd=False),
            os.path.join(subdir, subdir_name),
        )

        # empty dirs are skipped
        self._check_file(
            find_file(subdir_name, pathlist=['', self.tmpdir, subdir], cwd=False),
            os.path.join(subdir, subdir_name),
        )

    def test_find_library_system(self):
        # Find a system library (before we muck with the PATH)
        _args = {'cwd': False, 'include_PATH': False, 'pathlist': []}
        if FileDownloader.get_sysinfo()[0] == 'windows':
            a = find_library('ntdll', **_args)
            b = find_library('ntdll.dll', **_args)
            c = find_library('foo\\bar\\ntdll.dll', **_args)
        else:
            a = find_library('c', **_args)
            b = find_library('libc.so', **_args)
            c = find_library('foo/bar/libc.so', **_args)
        self.assertIsNotNone(a)
        self.assertIsNotNone(b)
        self.assertIsNotNone(c)
        self.assertEqual(a, b)
        # find_library could have found libc.so.6
        self.assertTrue(c.startswith(a))
        # Verify that the library is loadable (they are all the same
        # file, so only check one)
        _lib = ctypes.cdll.LoadLibrary(a)
        self.assertIsNotNone(_lib)

    def test_find_library_user(self):
        self.tmpdir = os.path.abspath(tempfile.mkdtemp())
        # CWD restored in tearDown
        os.chdir(self.tmpdir)

        envvar.PYOMO_CONFIG_DIR = self.tmpdir
        config_libdir = os.path.join(self.tmpdir, 'lib')
        os.mkdir(config_libdir)
        config_bindir = os.path.join(self.tmpdir, 'bin')
        os.mkdir(config_bindir)

        ldlibdir_name = 'in_ld_lib'
        ldlibdir = os.path.join(self.tmpdir, ldlibdir_name)
        os.mkdir(ldlibdir)
        os.environ['LD_LIBRARY_PATH'] = os.pathsep + ldlibdir + os.pathsep

        pathdir_name = 'in_path'
        pathdir = os.path.join(self.tmpdir, pathdir_name)
        os.mkdir(pathdir)
        os.environ['PATH'] = os.pathsep + pathdir + os.pathsep

        libExt = _libExt[_system()][0]

        f_in_cwd_ldlib_path = 'f_in_cwd_ldlib_path'
        open(os.path.join(self.tmpdir, f_in_cwd_ldlib_path), 'w').close()
        open(os.path.join(ldlibdir, f_in_cwd_ldlib_path), 'w').close()
        open(os.path.join(pathdir, f_in_cwd_ldlib_path), 'w').close()
        f_in_ldlib_extension = 'f_in_ldlib_extension'
        open(os.path.join(ldlibdir, f_in_ldlib_extension + libExt), 'w').close()
        f_in_path = 'f_in_path'
        open(os.path.join(pathdir, f_in_path), 'w').close()

        f_in_configlib = 'f_in_configlib'
        open(os.path.join(config_libdir, f_in_configlib), 'w').close()
        f_in_configbin = 'f_in_configbin'
        open(os.path.join(config_bindir, f_in_ldlib_extension), 'w').close()
        open(os.path.join(config_bindir, f_in_configbin), 'w').close()

        self._check_file(
            find_library(f_in_cwd_ldlib_path),
            os.path.join(self.tmpdir, f_in_cwd_ldlib_path),
        )
        self._check_file(
            os.path.join(ldlibdir, f_in_cwd_ldlib_path),
            find_library(f_in_cwd_ldlib_path, cwd=False),
        )
        self._check_file(
            os.path.join(ldlibdir, f_in_ldlib_extension) + libExt,
            find_library(f_in_ldlib_extension),
        )
        self._check_file(os.path.join(pathdir, f_in_path), find_library(f_in_path))
        if _system() == 'windows':
            self._check_file(
                os.path.join(pathdir, f_in_path),
                find_library(f_in_path, include_PATH=False),
            )
        else:
            # Note that on Windows, ctypes.util.find_library *always*
            # searches the PATH
            self.assertIsNone(find_library(f_in_path, include_PATH=False))
        self._check_file(
            os.path.join(pathdir, f_in_path),
            find_library(f_in_path, pathlist=os.pathsep + pathdir + os.pathsep),
        )
        # test an explicit pathlist overrides LD_LIBRARY_PATH
        self._check_file(
            os.path.join(pathdir, f_in_cwd_ldlib_path),
            find_library(f_in_cwd_ldlib_path, cwd=False, pathlist=[pathdir]),
        )
        # test that the PYOMO_CONFIG_DIR 'lib' dir is included
        self._check_file(
            os.path.join(config_libdir, f_in_configlib), find_library(f_in_configlib)
        )
        # and the Bin dir
        self._check_file(
            os.path.join(config_bindir, f_in_configbin), find_library(f_in_configbin)
        )
        # ... but only if include_PATH is true
        self.assertIsNone(find_library(f_in_configbin, include_PATH=False))
        # And none of them if the pathlist is specified
        self.assertIsNone(find_library(f_in_configlib, pathlist=pathdir))
        self.assertIsNone(find_library(f_in_configbin, pathlist=pathdir))

    def test_find_executable(self):
        self.tmpdir = os.path.abspath(tempfile.mkdtemp())
        # CWD restored in tearDown
        os.chdir(self.tmpdir)

        envvar.PYOMO_CONFIG_DIR = self.tmpdir
        config_libdir = os.path.join(self.tmpdir, 'lib')
        os.mkdir(config_libdir)
        config_bindir = os.path.join(self.tmpdir, 'bin')
        os.mkdir(config_bindir)

        ldlibdir_name = 'in_ld_lib'
        ldlibdir = os.path.join(self.tmpdir, ldlibdir_name)
        os.mkdir(ldlibdir)
        os.environ['LD_LIBRARY_PATH'] = os.pathsep + ldlibdir + os.pathsep

        pathdir_name = 'in_path'
        pathdir = os.path.join(self.tmpdir, pathdir_name)
        os.mkdir(pathdir)
        os.environ['PATH'] = os.pathsep + pathdir + os.pathsep

        exeExt = _exeExt[_system()] or ''

        f_in_cwd_notexe = 'f_in_cwd_notexe'
        open(os.path.join(self.tmpdir, f_in_cwd_notexe), 'w').close()
        f_in_cwd_ldlib_path = 'f_in_cwd_ldlib_path'
        self._make_exec(os.path.join(self.tmpdir, f_in_cwd_ldlib_path))
        self._make_exec(os.path.join(ldlibdir, f_in_cwd_ldlib_path))
        self._make_exec(os.path.join(pathdir, f_in_cwd_ldlib_path))
        f_in_path_extension = 'f_in_path_extension'
        self._make_exec(os.path.join(pathdir, f_in_path_extension + exeExt))
        f_in_path = 'f_in_path'
        self._make_exec(os.path.join(pathdir, f_in_path))

        f_in_configlib = 'f_in_configlib'
        self._make_exec(os.path.join(config_libdir, f_in_configlib))
        f_in_configbin = 'f_in_configbin'
        self._make_exec(os.path.join(config_libdir, f_in_path_extension))
        self._make_exec(os.path.join(config_bindir, f_in_configbin))

        found = find_executable(f_in_cwd_notexe)
        if _system() in ('windows', 'cygwin'):
            self._check_file(found, os.path.join(self.tmpdir, f_in_cwd_notexe))
        else:
            self.assertIsNone(found)

        self._check_file(
            find_executable(f_in_cwd_ldlib_path),
            os.path.join(self.tmpdir, f_in_cwd_ldlib_path),
        )
        self._check_file(
            os.path.join(pathdir, f_in_cwd_ldlib_path),
            find_executable(f_in_cwd_ldlib_path, cwd=False),
        )
        self._check_file(
            find_executable(f_in_path_extension),
            os.path.join(pathdir, f_in_path_extension) + exeExt,
        )
        self._check_file(find_executable(f_in_path), os.path.join(pathdir, f_in_path))
        self.assertIsNone(find_executable(f_in_path, include_PATH=False))
        self._check_file(
            find_executable(f_in_path, pathlist=os.pathsep + pathdir + os.pathsep),
            os.path.join(pathdir, f_in_path),
        )

        # test an explicit pathlist overrides PATH
        self._check_file(
            os.path.join(ldlibdir, f_in_cwd_ldlib_path),
            find_executable(f_in_cwd_ldlib_path, cwd=False, pathlist=[ldlibdir]),
        )
        # test that the PYOMO_CONFIG_DIR 'bin' dir is included
        self._check_file(
            os.path.join(config_bindir, f_in_configbin), find_executable(f_in_configbin)
        )
        # ... but only if the pathlist is not specified
        self.assertIsNone(find_executable(f_in_configbin, pathlist=pathdir))

    def test_PathManager(self):
        Executable = PathManager(find_executable, ExecutableData)
        self.tmpdir = os.path.abspath(tempfile.mkdtemp())

        envvar.PYOMO_CONFIG_DIR = self.tmpdir
        config_bindir = os.path.join(self.tmpdir, 'bin')
        os.mkdir(config_bindir)

        pathdir_name = 'in_path'
        pathdir = os.path.join(self.tmpdir, pathdir_name)
        os.mkdir(pathdir)
        os.environ['PATH'] = os.pathsep + pathdir + os.pathsep

        f_in_tmp = 'f_in_tmp'
        self._make_exec(os.path.join(self.tmpdir, f_in_tmp))
        f_in_path = 'f_in_path'
        self._make_exec(os.path.join(pathdir, f_in_path))
        f_in_cfg = 'f_in_configbin'
        self._make_exec(os.path.join(config_bindir, f_in_cfg))

        # Test availability
        self.assertTrue(Executable(f_in_path).available())
        if not Executable(f_in_path):
            self.fail("Expected casting Executable(f_in_path) to bool=True")

        # Test getting the path to the executable
        self._check_file(Executable(f_in_path).path(), os.path.join(pathdir, f_in_path))
        self._check_file("%s" % Executable(f_in_path), os.path.join(pathdir, f_in_path))
        self._check_file(
            Executable(f_in_path).executable, os.path.join(pathdir, f_in_path)
        )

        # Test the above for a nonexistent file
        self.assertFalse(Executable(f_in_tmp).available())
        if Executable(f_in_tmp):
            self.fail("Expected casting Executable(f_in_tmp) to bool=False")
        self.assertIsNone(Executable(f_in_tmp).path())
        self.assertEqual("%s" % Executable(f_in_tmp), "")
        self.assertIsNone(Executable(f_in_tmp).executable)

        # If we override the pathlist, then we will not find the CONFIGDIR
        Executable.pathlist = []
        self.assertFalse(Executable(f_in_cfg).available())
        Executable.pathlist.append(config_bindir)
        # and adding it won't change things (status is cached)
        self.assertFalse(Executable(f_in_cfg).available())
        # until we tell the manager to rehash the executables
        Executable.rehash()
        self.assertTrue(Executable(f_in_cfg).available())
        self.assertEqual(
            Executable(f_in_cfg).path(), os.path.join(config_bindir, f_in_cfg)
        )
        # Note that if we clear the pathlist, then the current value of
        # CONFIGDIR will be honored
        Executable.pathlist = None
        Executable.rehash()
        self.assertTrue(Executable(f_in_cfg).available())
        self.assertEqual(
            Executable(f_in_cfg).path(), os.path.join(config_bindir, f_in_cfg)
        )

        # Another file that doesn't exist
        f_in_path2 = 'f_in_path2'
        f_loc = os.path.join(pathdir, f_in_path2)
        self.assertFalse(Executable(f_in_path2).available())
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.common', logging.WARNING):
            Executable(f_in_path2).executable = f_loc
            self.assertIn(
                "explicitly setting the path for '%s' to an "
                "invalid object or nonexistent location ('%s')" % (f_in_path2, f_loc),
                output.getvalue(),
            )
        self.assertFalse(Executable(f_in_path2).available())
        self._make_exec(os.path.join(pathdir, f_in_path2))
        self.assertFalse(Executable(f_in_path2).available())
        Executable(f_in_path2).rehash()
        self.assertTrue(Executable(f_in_path2).available())

        # And disabling it will "remove" it
        Executable(f_in_path2).disable()
        self.assertFalse(Executable(f_in_path2).available())
        self.assertIsNone(Executable(f_in_path2).path())
        Executable(f_in_path2).rehash()
        self.assertTrue(Executable(f_in_path2).available())
        self.assertEqual(Executable(f_in_path2).path(), f_loc)

    def test_to_legal_filename(self):
        self.assertEqual('abc', to_legal_filename('abc'))
        self.assertEqual('', to_legal_filename(''))
        if envvar.is_windows:
            self.assertEqual('_abc', to_legal_filename(' abc'))
            self.assertEqual('abc_', to_legal_filename('abc.'))
            self.assertEqual('abc_', to_legal_filename('abc '))
            self.assertEqual('abc_def', to_legal_filename('abc/def'))
            self.assertEqual('abc_def', to_legal_filename('abc\\def'))
            self.assertEqual(
                'a_b_c', to_legal_filename(''.join(['a', chr(0), 'b', chr(7), 'c']))
            )
        else:
            self.assertEqual(' abc', to_legal_filename(' abc'))
            self.assertEqual('abc.', to_legal_filename('abc.'))
            self.assertEqual('abc ', to_legal_filename('abc '))
            self.assertEqual('abc_def', to_legal_filename('abc/def'))
            self.assertEqual('abc\\def', to_legal_filename('abc\\def'))
            self.assertEqual(
                'a_b' + chr(7) + 'c',
                to_legal_filename(''.join(['a', chr(0), 'b', chr(7), 'c'])),
            )

        self.assertEqual('_abc', to_legal_filename(' abc', True))
        self.assertEqual('abc_', to_legal_filename('abc.', True))
        self.assertEqual('abc_', to_legal_filename('abc ', True))
        self.assertEqual('abc_def', to_legal_filename('abc/def', True))
        self.assertEqual('abc_def', to_legal_filename('abc\\def', True))
        self.assertEqual(
            'a_b_c', to_legal_filename(''.join(['a', chr(0), 'b', chr(7), 'c']), True)
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import logging
from weakref import ref as weakref_ref

from pyomo.common.deprecation import RenamedClass, deprecated
from pyomo.common.formatting import tabular_writer
from pyomo.common.log import is_debug_set
from pyomo.common.modeling import NOTSET
from pyomo.common.pyomo_typing import overload
from pyomo.common.timing import ConstructionTimer

from pyomo.core.expr.expr_common import _type_check_exception_arg
from pyomo.core.expr.boolean_value import as_boolean, BooleanConstant
from pyomo.core.expr.numvalue import native_types, native_logical_types
from pyomo.core.base.component import ActiveComponentData, ModelComponentFactory
from pyomo.core.base.disable_methods import disable_methods
from pyomo.core.base.global_set import UnindexedComponent_index
from pyomo.core.base.indexed_component import (
    ActiveIndexedComponent,
    UnindexedComponent_set,
)
from pyomo.core.base.initializer import Initializer
from pyomo.core.base.set import Set

logger = logging.getLogger(__name__)
_known_logical_expression_types = set()
_rule_returned_none_error = """LogicalConstraint '%s': rule returned None.

logical constraint rules must return a valid logical proposition.
The most common cause of this error is
forgetting to include the "return" statement at the end of your rule.
"""


class LogicalConstraintData(ActiveComponentData):
    """
    This class defines the data for a single general logical constraint.

    Constructor arguments:
        component
            The LogicalConstraint object that owns this data.
        expr
            The Pyomo expression stored in this logical constraint.

    Public class attributes:
        active
            A boolean that is true if this logical constraint is
            active in the model.
        expr
            The Pyomo expression for this logical constraint

    Private class attributes:
        _component
            The logical constraint component.
        _active
            A boolean that indicates whether this data is active

    """

    __slots__ = ('_expr',)

    def __init__(self, expr=None, component=None):
        #
        # These lines represent in-lining of the
        # following constructors:
        #   - LogicalConstraintData,
        #   - ActiveComponentData
        #   - ComponentData
        self._component = weakref_ref(component) if (component is not None) else None
        self._index = NOTSET
        self._active = True

        self._expr = None
        if expr is not None:
            self.set_value(expr)

    def __call__(self, exception=NOTSET):
        """Compute the value of the body of this logical constraint."""
        exception = _type_check_exception_arg(self, exception)
        return self.expr(exception=exception)

    #
    # Abstract Interface
    #

    @property
    def body(self):
        """Access the body of a logical constraint expression."""
        return self.expr

    @property
    def expr(self):
        """Return the expression associated with this logical constraint."""
        return self._expr

    def set_value(self, expr):
        """Set the expression on this logical constraint."""
        if expr.__class__ in _known_logical_expression_types:
            self._expr = expr
            return
        #
        # Ignore an 'empty' constraint
        #
        if expr is LogicalConstraint.Skip:
            del self.parent_component()[self.index()]
            return

        elif expr.__class__ in native_logical_types:
            self._expr = as_boolean(expr)
            return

        elif expr is None:
            raise ValueError(_rule_returned_none_error % (self.name,))

        elif expr.__class__ not in native_types:
            try:
                if expr.is_logical_type():
                    self._expr = expr
                    _known_logical_expression_types.add(expr.__class__)
                    return

                # FIXME: we should extend the templating system to
                # handle things like IntervalVar types so that
                # indirection and things like CallExpression can
                # properly propagate the expression type.  In the
                # meantime, we will just assume that all template
                # expressions are acceptable (which, while not correct,
                # is consistent with prior behavior)
                if hasattr(expr, '_resolve_template'):
                    self._expr = expr
                    return
            except (AttributeError, TypeError):
                pass

        raise ValueError(
            "Assigning improper value to LogicalConstraint '%s'. "
            "Found %s '%s'.\n"
            "Expecting a logical expression or Boolean value. Examples:"
            "\n   (m.Y1 & m.Y2).implies(m.Y3)"
            "\n   atleast(1, m.Y1, m.Y2)" % (self.name, type(expr).__name__, str(expr))
        )

    def get_value(self):
        """Get the expression on this logical constraint."""
        return self.expr


class _LogicalConstraintData(metaclass=RenamedClass):
    __renamed__new_class__ = LogicalConstraintData
    __renamed__version__ = '6.7.2'


class _GeneralLogicalConstraintData(metaclass=RenamedClass):
    __renamed__new_class__ = LogicalConstraintData
    __renamed__version__ = '6.7.2'


@ModelComponentFactory.register("General logical constraints.")
class LogicalConstraint(ActiveIndexedComponent):
    """
    This modeling component defines a logical constraint using a
    rule function.

    Constructor arguments:
        expr
            A Pyomo expression for this logical constraint
        rule
            A function that is used to construct logical constraints
        doc
            A text string describing this component
        name
            A name for this component

    Public class attributes:
        doc
            A text string describing this component
        name
            A name for this component
        active
            A boolean that is true if this component will be used to
            construct a model instance
        rule
           The rule used to initialize the logical constraint(s)

    Private class attributes:
        _constructed
            A boolean that is true if this component has been constructed
        _data
            A dictionary from the index set to component data objects
        _index_set
            The set of valid indices
        _model
            A weakref to the model that owns this component
        _parent
            A weakref to the parent block that owns this component
        _type
            The class type for the derived subclass
    """

    _ComponentDataClass = LogicalConstraintData

    Infeasible = BooleanConstant(False, 'Infeasible')
    Feasible = BooleanConstant(True, 'Feasible')

    NoConstraint = ActiveIndexedComponent.Skip
    Violated = Infeasible
    Satisfied = Feasible

    def __new__(cls, *args, **kwds):
        if cls != LogicalConstraint:
            return super().__new__(cls)
        if not args or (args[0] is UnindexedComponent_set and len(args) == 1):
            return super().__new__(AbstractScalarLogicalConstraint)
        else:
            return super().__new__(IndexedLogicalConstraint)

    def __init__(self, *args, **kwargs):
        _init = self._pop_from_kwargs('Constraint', kwargs, ('rule', 'expr'), None)
        self._rule = Initializer(_init)

        kwargs.setdefault('ctype', LogicalConstraint)
        ActiveIndexedComponent.__init__(self, *args, **kwargs)

    def construct(self, data=None):
        """
        Construct the expression(s) for this logical constraint.
        """
        if self._constructed:
            return
        self._constructed = True

        timer = ConstructionTimer(self)
        if is_debug_set(logger):
            logger.debug("Constructing logical constraint %s" % self.name)

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        rule = self._rule
        try:
            # We do not (currently) accept data for constructing LogicalConstraints
            index = None
            assert data is None

            if rule is None:
                # If there is no rule, then we are immediately done.
                return

            if rule.constant() and self.is_indexed():
                raise IndexError(
                    "LogicalConstraint '%s': Cannot initialize multiple indices "
                    "of a constraint with a single expression" % (self.name,)
                )

            block = self.parent_block()
            if rule.contains_indices():
                # The index is coming in externally; we need to validate it
                for index in rule.indices():
                    self[index] = rule(block, index)
            elif not self.index_set().isfinite():
                # If the index is not finite, then we cannot iterate
                # over it.  Since the rule doesn't provide explicit
                # indices, then there is nothing we can do (the
                # assumption is that the user will trigger specific
                # indices to be created at a later time).
                pass
            else:
                # Bypass the index validation and create the member directly
                for index in self.index_set():
                    self._setitem_when_not_present(index, rule(block, index))
        except Exception:
            err = sys.exc_info()[1]
            logger.error(
                "Rule failed when generating expression for "
                "LogicalConstraint %s with index %s:\n%s: %s"
                % (self.name, str(index), type(err).__name__, err)
            )
            raise
        finally:
            timer.report()

    def _getitem_when_not_present(self, idx):
        if self._rule is None:
            raise KeyError(idx)
        con = self._setitem_when_not_present(idx, self._rule(self.parent_block(), idx))
        if con is None:
            raise KeyError(idx)
        return con

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        return (
            [
                ("Size", len(self)),
                ("Index", self._index_set if self.is_indexed() else None),
                ("Active", self.active),
            ],
            self.items,
            ("Body", "Active"),
            lambda k, v: [v.body, v.active],
        )

    @property
    def rule(self):
        return self._rule

    @rule.setter
    @deprecated(
        f"The 'LogicalConstraint.rule' attribute will be made read-only",
        version='6.9.3',
        remove_in='6.11',
    )
    def rule(self, rule):
        self._rule = rule

    def display(self, prefix="", ostream=None):
        """
        Print component state information

        This duplicates logic in Component.pprint()
        """
        if not self.active:
            return
        if ostream is None:
            ostream = sys.stdout
        tab = "    "
        ostream.write(prefix + self.local_name + " : ")
        ostream.write("Size=" + str(len(self)))

        ostream.write("\n")
        tabular_writer(
            ostream,
            prefix + tab,
            ((k, v) for k, v in self._data.items() if v.active),
            ("Body",),
            lambda k, v: [v.body()],
        )


class ScalarLogicalConstraint(LogicalConstraintData, LogicalConstraint):
    """
    ScalarLogicalConstraint is the implementation representing a single,
    non-indexed logical constraint.
    """

    def __init__(self, *args, **kwds):
        LogicalConstraintData.__init__(self, component=self, expr=None)
        LogicalConstraint.__init__(self, *args, **kwds)
        self._index = UnindexedComponent_index

    #
    # Override abstract interface methods to first check for
    # construction
    #

    @property
    def expr(self):
        """Access the body of a logical constraint."""
        if not self._data:
            raise ValueError(
                "Accessing the expr of ScalarLogicalConstraint "
                "'%s' before the LogicalConstraint has been assigned "
                "an expression. There is currently "
                "nothing to access." % self.name
            )
        return LogicalConstraintData.expr.fget(self)

    #
    # Singleton logical constraints are strange in that we want them to be
    # both be constructed but have len() == 0 when not initialized with
    # anything (at least according to the unit tests that are
    # currently in place). So during initialization only, we will
    # treat them as "indexed" objects where things like
    # True are managed. But after that they will behave
    # like LogicalConstraintData objects where set_value expects
    # a valid expression or None.
    #

    def set_value(self, expr):
        """Set the expression on this logical constraint."""
        if not self._data:
            self._data[None] = self
        return super().set_value(expr)

    #
    # Leaving this method for backward compatibility reasons.
    # (probably should be removed)
    #
    def add(self, index, expr):
        """Add a logical constraint with a given index."""
        if index is not None:
            raise ValueError(
                "ScalarLogicalConstraint object '%s' does not accept "
                "index values other than None. Invalid value: %s" % (self.name, index)
            )
        self.set_value(expr)
        return self


class SimpleLogicalConstraint(metaclass=RenamedClass):
    __renamed__new_class__ = ScalarLogicalConstraint
    __renamed__version__ = '6.0'


@disable_methods({'add', 'set_value', 'expr', 'body'})
class AbstractScalarLogicalConstraint(ScalarLogicalConstraint):
    pass


class IndexedLogicalConstraint(LogicalConstraint):
    #
    # Leaving this method for backward compatibility reasons
    #
    # Note: Beginning after Pyomo 5.2 this method will now validate that
    # the index is in the underlying index set (through 5.2 the index
    # was not checked).
    #
    def add(self, index, expr):
        """Add a logical constraint with a given index."""
        return self.__setitem__(index, expr)

    @overload
    def __getitem__(self, index) -> LogicalConstraintData: ...

    __getitem__ = ActiveIndexedComponent.__getitem__  # type: ignore


@ModelComponentFactory.register("A list of logical constraints.")
class LogicalConstraintList(IndexedLogicalConstraint):
    """
    A logical constraint component that represents a list of constraints.
    Constraints can be indexed by their index, but when they are
    added an index value is not specified.
    """

    class End:
        pass

    def __init__(self, **kwargs):
        """Constructor"""
        if 'expr' in kwargs:
            raise ValueError("LogicalConstraintList does not accept the 'expr' keyword")
        _rule = kwargs.pop('rule', None)
        self._starting_index = kwargs.pop('starting_index', 1)

        super().__init__(Set(dimen=1), **kwargs)

        self._rule = Initializer(
            _rule, treat_sequences_as_mappings=False, allow_generators=True
        )

    def construct(self, data=None):
        """
        Construct the expression(s) for this logical constraint.
        """
        if self._constructed:
            return
        self._constructed = True

        if is_debug_set(logger):
            logger.debug("Constructing logical constraint list %s" % (self.name))

        if self._anonymous_sets is not None:
            for _set in self._anonymous_sets:
                _set.construct()

        if self._rule is not None:
            _rule = self._rule(self.parent_block(), ())
            for cc in iter(_rule):
                if cc is LogicalConstraintList.End:
                    break
                if cc is LogicalConstraint.Skip:
                    continue
                self.add(cc)

    def add(self, expr):
        """Add a logical constraint with an implicit index."""
        next_idx = len(self._index_set) + self._starting_index
        self._index_set.add(next_idx)
        return self.__setitem__(next_idx, expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import re
import copy
import logging

from pyomo.common.log import is_debug_set
from pyomo.common.collections import Bunch, OrderedDict
from pyomo.common.errors import ApplicationError

from pyomo.dataportal.parse_datacmds import parse_data_commands, _re_number
from pyomo.dataportal.factory import DataManagerFactory, UnknownDataManager
from pyomo.core.base.set import UnknownSetDimen
from pyomo.core.base.util import flatten_tuple

numlist = {bool, int, float}

logger = logging.getLogger('pyomo.core')

global Lineno
global Filename

_num_pattern = re.compile("^(" + _re_number + ")$")
_str_false_values = {'False', 'false', 'FALSE'}
_str_bool_values = {'True', 'true', 'TRUE'}
_str_bool_values.update(_str_false_values)


def _guess_set_dimen(index):
    d = 0
    # Look through the subsets of this index and get their dimen
    for subset in index.subsets():
        sub_d = subset.dimen
        # If the subset has an unknown dimen, then look at the subset's
        # domain to guess the dimen.
        if sub_d is UnknownSetDimen:
            for domain_subset in subset.domain.subsets():
                sub_d = domain_subset.domain.dimen
                if sub_d in (UnknownSetDimen, None):
                    # We will guess that None / Unknown domains are dimen==1
                    d += 1
                else:
                    d += sub_d
        elif sub_d is None:
            return None
        else:
            d += sub_d
    return d


def _process_token(token):
    # print("TOKEN:", token, type(token))
    if type(token) is tuple:
        return tuple(_process_token(i) for i in token)
    elif type(token) in numlist:
        return token
    elif token in _str_bool_values:
        return token not in _str_false_values
    elif token[0] == '"' and token[-1] == '"':
        # Strip "flag" quotation characters
        return token[1:-1]
    elif token[0] == '[' and token[-1] == ']':
        vals = []
        token = token[1:-1]
        for item in token.split(","):
            if item[0] in '"\'' and item[0] == item[-1]:
                vals.append(item[1:-1])
            elif _num_pattern.match(item):
                _num = float(item)
                if '.' in item:
                    vals.append(_num)
                else:
                    _int = int(_num)
                    vals.append(_int if _int == _num else _num)
            else:
                vals.append(item)
        return tuple(vals)
    elif _num_pattern.match(token):
        _num = float(token)
        if '.' in token:
            return _num
        else:
            _int = int(_num)
            return _int if _int == _num else _num
    else:
        return token


def _preprocess_data(cmd):
    """
    Called by _process_data() to (1) combine tokens that comprise a tuple
    and (2) combine the ':' token with the previous token
    """
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("_preprocess_data(start) %s", cmd)
    state = 0
    newcmd = []
    tpl = []
    for token in cmd:
        if state == 0:
            if type(token) in numlist:
                newcmd.append(token)
            elif token == ',':
                raise ValueError(
                    "Unexpected comma outside of (), {} or [] declarations"
                )
            elif token == '(':
                state = 1
            elif token == ')':
                raise ValueError("Unexpected ')' that does not follow a '('")
            elif token == '{':
                state = 2
            elif token == '}':
                raise ValueError("Unexpected '}' that does not follow a '{'")
            elif token == '[':
                state = 3
            elif token == ']':
                raise ValueError("Unexpected ']' that does not follow a '['")
            else:
                newcmd.append(_process_token(token))

        elif state == 1:
            # After a '('
            if type(token) in numlist:
                tpl.append(token)
            elif token == ',':
                pass
            elif token == '(':
                raise ValueError("Two '('s follow each other in the data")
            elif token == ')':
                newcmd.append(tuple(tpl))
                tpl = []
                state = 0
            else:
                tpl.append(_process_token(token))

        elif state == 2:
            # After a '{'
            if type(token) in numlist:
                tpl.append(token)
            elif token == ',':
                pass
            elif token == '{':
                raise ValueError("Two '{'s follow each other in the data")
            elif token == '}':
                newcmd.append(
                    tpl
                )  # Keep this as a list, so we can distinguish it while parsing tables
                tpl = []
                state = 0
            else:
                tpl.append(_process_token(token))

        elif state == 3:
            # After a '['
            if type(token) in numlist:
                tpl.append(token)
            elif token == ',':
                pass
            elif token == '[':
                raise ValueError("Two '['s follow each other in the data")
            elif token == ']':
                newcmd.append(tuple(tpl))
                tpl = []
                state = 0
            else:
                tpl.append(_process_token(token))

    if state == 1:
        raise ValueError("Data ends without tuple ending")
    elif state == 2:
        raise ValueError("Data ends without braces ending")
    elif state == 3:
        raise ValueError("Data ends without bracket ending")
    if generate_debug_messages:
        logger.debug("_preprocess_data(end) %s", newcmd)
    return newcmd


def _process_set(cmd, _model, _data):
    """
    Called by _process_data() to process a set declaration.
    """
    # print("SET %s" % cmd)
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("DEBUG: _process_set(start) %s", cmd)
    #
    # Process a set
    #
    if type(cmd[2]) is tuple:
        #
        # An indexed set
        #
        ndx = cmd[2]
        if len(ndx) == 0:
            # At this point, if the index is an empty tuple, then there is an
            # issue with the specification of this indexed set.
            raise ValueError(
                "Illegal indexed set specification encountered: " + str(cmd[1])
            )
        elif len(ndx) == 1:
            ndx = ndx[0]
        if cmd[1] not in _data:
            _data[cmd[1]] = {}
        _data[cmd[1]][ndx] = _process_set_data(cmd[4:], cmd[1], _model)

    elif cmd[2] == ":":
        #
        # A tabular set
        #
        _data[cmd[1]] = {}
        _data[cmd[1]][None] = []
        i = 3
        while cmd[i] != ":=":
            i += 1
        ndx1 = cmd[3:i]
        i += 1
        while i < len(cmd):
            ndx = cmd[i]
            for j in range(0, len(ndx1)):
                if cmd[i + j + 1] == "+":
                    # print("DATA %s %s" % (ndx1[j], cmd[i]))
                    _data[cmd[1]][None].append((ndx1[j], cmd[i]))
            i += len(ndx1) + 1
    else:
        #
        # Processing a general set
        #
        _data[cmd[1]] = {}
        _data[cmd[1]][None] = _process_set_data(cmd[3:], cmd[1], _model)


def _process_set_data(cmd, sname, _model):
    """
    Called by _process_set() to process set data.
    """
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("DEBUG: _process_set_data(start) %s", cmd)
    if len(cmd) == 0:
        return []
    ans = []
    i = 0
    template = None
    ndx = []
    template = []
    while i < len(cmd):
        if type(cmd[i]) is not tuple:
            if len(ndx) == 0:
                ans.append(cmd[i])
            else:
                #
                # Use the ndx to create a tuple.  This list
                # contains the indices of the values that need
                # to be filled-in
                #
                tmpval = template
                for kk in range(len(ndx)):
                    if i == len(cmd):
                        raise IOError(
                            "Expected another set value to flush out a tuple pattern!"
                        )
                    tmpval[ndx[kk]] = cmd[i]
                    i += 1
                ans.append(tuple(tmpval))
                continue
        elif "*" not in cmd[i]:
            ans.append(cmd[i])
        else:
            template = list(cmd[i])
            ndx = []
            for kk in range(len(template)):
                if template[kk] == '*':
                    ndx.append(kk)
        i += 1
    if generate_debug_messages:
        logger.debug("DEBUG: _process_set_data(end) %s", ans)
    return ans


def _process_param(cmd, _model, _data, _default, index=None, param=None, ncolumns=None):
    """
    Called by _process_data to process data for a Parameter declaration
    """
    # print('PARAM %s index=%s ncolumns=%s' %(cmd, index, ncolumns))
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("DEBUG: _process_param(start) %s", cmd)
    #
    # Process parameters
    #
    dflt = None
    singledef = True
    cmd = cmd[1:]
    if cmd[0] == ":":
        singledef = False
        cmd = cmd[1:]
    # print "SINGLEDEF", singledef
    if singledef:
        pname = cmd[0]
        cmd = cmd[1:]
        if len(cmd) >= 2 and cmd[0] == "default":
            dflt = cmd[1]
            cmd = cmd[2:]
        if dflt != None:
            _default[pname] = dflt
        if cmd[0] == ":=":
            cmd = cmd[1:]
        transpose = False
        if cmd[0] == "(tr)":
            transpose = True
            if cmd[1] == ":":
                cmd = cmd[1:]
            else:
                cmd[0] = ":"
        if cmd[0] != ":":
            # print "HERE YYY", pname, transpose, _model, ncolumns
            if generate_debug_messages:
                logger.debug("DEBUG: _process_param (singledef without :...:=) %s", cmd)
            cmd = _apply_templates(cmd)
            # print 'cmd',cmd
            if not transpose:
                if pname not in _data:
                    _data[pname] = {}
                if not ncolumns is None:
                    finaldata = _process_data_list(pname, ncolumns - 1, cmd)
                elif not _model is None:
                    _param = getattr(_model, pname)
                    _dim = _param.dim()
                    if _dim is UnknownSetDimen:
                        _dim = _guess_set_dimen(_param.index_set())
                    finaldata = _process_data_list(pname, _dim, cmd)
                else:
                    _dim = 1 if len(cmd) > 1 else 0
                    finaldata = _process_data_list(pname, _dim, cmd)
                for key in finaldata:
                    _data[pname][key] = finaldata[key]
            else:
                tmp = ["param", pname, ":="]
                i = 1
                while i < len(cmd):
                    i0 = i
                    while cmd[i] != ":=":
                        i = i + 1
                    ncol = i - i0 + 1
                    lcmd = i
                    while lcmd < len(cmd) and cmd[lcmd] != ":":
                        lcmd += 1
                    j0 = i0 - 1
                    for j in range(1, ncol):
                        ii = 1 + i
                        kk = ii + j
                        while kk < lcmd:
                            if cmd[kk] != ".":
                                # if 1>0:
                                tmp.append(copy.copy(cmd[j + j0]))
                                tmp.append(copy.copy(cmd[ii]))
                                tmp.append(copy.copy(cmd[kk]))
                            ii = ii + ncol
                            kk = kk + ncol
                    i = lcmd + 1
                _process_param(
                    tmp,
                    _model,
                    _data,
                    _default,
                    index=index,
                    param=param,
                    ncolumns=ncolumns,
                )
        else:
            tmp = ["param", pname, ":="]
            if param is None:
                param = [pname]
            i = 1
            if generate_debug_messages:
                logger.debug("DEBUG: _process_param (singledef with :...:=) %s", cmd)
            while i < len(cmd):
                i0 = i
                while i < len(cmd) and cmd[i] != ":=":
                    i = i + 1
                if i == len(cmd):
                    raise ValueError(
                        "ERROR: Trouble on line " + str(Lineno) + " of file " + Filename
                    )
                ncol = i - i0 + 1
                lcmd = i
                while lcmd < len(cmd) and cmd[lcmd] != ":":
                    lcmd += 1
                j0 = i0 - 1
                for j in range(1, ncol):
                    ii = 1 + i
                    kk = ii + j
                    while kk < lcmd:
                        if cmd[kk] != ".":
                            if transpose:
                                tmp.append(copy.copy(cmd[j + j0]))
                                tmp.append(copy.copy(cmd[ii]))
                            else:
                                tmp.append(copy.copy(cmd[ii]))
                                tmp.append(copy.copy(cmd[j + j0]))
                            tmp.append(copy.copy(cmd[kk]))
                        ii = ii + ncol
                        kk = kk + ncol
                i = lcmd + 1
                _process_param(
                    tmp,
                    _model,
                    _data,
                    _default,
                    index=index,
                    param=param[0],
                    ncolumns=3,
                )

    else:
        if generate_debug_messages:
            logger.debug("DEBUG: _process_param (cmd[0]=='param:') %s", cmd)
        i = 0
        nsets = 0
        while i < len(cmd):
            if cmd[i] == ':=':
                i = -1
                break
            if cmd[i] == ":":
                nsets = i
                break
            i += 1
        nparams = 0
        _i = i + 1
        while i < len(cmd):
            if cmd[i] == ':=':
                nparams = i - _i
                break
            i += 1
        if i == len(cmd):
            raise ValueError(
                "Trouble on data file line " + str(Lineno) + " of file " + Filename
            )
        if generate_debug_messages:
            logger.debug("NSets %d", nsets)
        Lcmd = len(cmd)
        j = 0
        d = 1
        # print "HERE", nsets, nparams
        #
        # Process sets first
        #
        while j < nsets:
            # NOTE: I'm pretty sure that nsets is always equal to 1
            sname = cmd[j]
            if not ncolumns is None:
                d = ncolumns - nparams
            elif _model is None:
                d = 1
            else:
                index = getattr(_model, sname)
                d = _guess_set_dimen(index)
            # print "SET",sname,d,_model#,getattr(_model,sname).dimen, type(index)
            # d = getattr(_model,sname).dimen
            np = i - 1
            if generate_debug_messages:
                logger.debug("I %d, J %d, SName %s, d %d", i, j, sname, d)
            dnp = d + np - 1
            # k0 = i + d - 2
            ii = i + j + 1
            tmp = ["set", cmd[j], ":="]
            while ii < Lcmd:
                if d > 1:
                    _tmp = []
                    for dd in range(0, d):
                        _tmp.append(copy.copy(cmd[ii + dd]))
                    tmp.append(tuple(_tmp))
                else:
                    for dd in range(0, d):
                        tmp.append(copy.copy(cmd[ii + dd]))
                ii += dnp
            _process_set(tmp, _model, _data)
            j += 1
        if nsets > 0:
            j += 1
        #
        # Process parameters second
        #
        # print "HERE", cmd
        # print "HERE", param
        # print "JI",j,i # XXX
        jstart = j
        if param is None:
            param = []
            _j = j
            while _j < i:
                param.append(cmd[_j])
                _j += 1
        while j < i:
            # print "HERE", i, j, jstart, cmd[j]
            pname = param[j - jstart]
            if generate_debug_messages:
                logger.debug("I %d, J %d, Pname %s", i, j, pname)
            if not ncolumns is None:
                d = ncolumns - nparams
            elif _model is None:
                d = 1
            else:
                _param = getattr(_model, pname)
                d = _param.dim()
                if d is UnknownSetDimen:
                    d = _guess_set_dimen(_param.index_set())
            if nsets > 0:
                np = i - 1
                dnp = d + np - 1
                ii = i + 1
                kk = i + d + j - 1
            else:
                np = i
                dnp = d + np
                ii = i + 1
                kk = np + 1 + d + nsets + j
            # print cmd[ii], d, np, dnp, ii, kk
            tmp = ["param", pname, ":="]
            if generate_debug_messages:
                logger.debug('dnp %d\nnp %d', dnp, np)
            while kk < Lcmd:
                if generate_debug_messages:
                    logger.debug("kk %d, ii %d", kk, ii)
                iid = ii + d
                while ii < iid:
                    tmp.append(copy.copy(cmd[ii]))
                    ii += 1
                ii += dnp - d
                tmp.append(copy.copy(cmd[kk]))
                kk += dnp
            # print "TMP", tmp, ncolumns-nparams+1
            if not ncolumns is None:
                nc = ncolumns - nparams + 1
            else:
                nc = None
            _process_param(
                tmp,
                _model,
                _data,
                _default,
                index=index,
                param=param[j - jstart],
                ncolumns=nc,
            )
            j += 1


def _apply_templates(cmd):
    template = []
    ilist = set()
    ans = []
    i = 0
    while i < len(cmd):
        # print i, len(cmd), cmd[i], ilist, template, ans
        if type(cmd[i]) is tuple and '*' in cmd[i]:
            j = i
            tmp = list(cmd[j])
            nindex = len(tmp)
            template = tmp
            ilist = set()
            for kk in range(nindex):
                if tmp[kk] == '*':
                    ilist.add(kk)
        elif len(ilist) == 0:
            ans.append(cmd[i])
        else:
            for kk in range(len(template)):
                if kk in ilist:
                    ans.append(cmd[i])
                    i += 1
                else:
                    ans.append(template[kk])
            ans.append(cmd[i])
        i += 1
    return ans


def _process_data_list(param_name, dim, cmd):
    """\
 Called by _process_param() to process a list of data for a Parameter.
 """
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("process_data_list %d %s", dim, cmd)

    if len(cmd) % (dim + 1) != 0:
        msg = (
            "Parameter '%s' defined with '%d' dimensions, "
            "but data has '%d' values: %s."
        )
        msg = msg % (param_name, dim, len(cmd), cmd)
        if len(cmd) % (dim + 1) == dim:
            msg += " Are you missing a value for a %d-dimensional index?" % dim
        elif len(cmd) % dim == 0:
            msg += " Are you missing the values for %d-dimensional indices?" % dim
        else:
            msg += " Data declaration must be given in multiples of %d." % (dim + 1)
        raise ValueError(msg)

    ans = {}
    if dim == 0:
        ans[None] = cmd[0]
        return ans
    i = 0
    while i < len(cmd):
        if dim > 1:
            ndx = tuple(cmd[i : i + dim])
        else:
            ndx = cmd[i]
        if cmd[i + dim] != ".":
            ans[ndx] = cmd[i + dim]
        i += dim + 1
    return ans


def _process_include(cmd, _model, _data, _default, options=None):
    if len(cmd) == 1:
        raise IOError("Cannot execute 'include' command without a filename")
    if len(cmd) > 2:
        raise IOError("The 'include' command only accepts a single filename")

    global Filename
    Filename = cmd[1]
    global Lineno
    Lineno = 0
    try:
        scenarios = parse_data_commands(filename=cmd[1])
    except IOError:
        raise
        err = sys.exc_info()[1]
        raise IOError("Error parsing file '%s': %s" % (Filename, str(err)))
    if scenarios is None:
        return False
    for scenario in scenarios:
        for cmd in scenarios[scenario]:
            if scenario not in _data:
                _data[scenario] = {}
            if cmd[0] in ('include', 'load'):
                _tmpdata = {}
                _process_data(cmd, _model, _tmpdata, _default, Filename, Lineno)
                if scenario is None:
                    for key in _tmpdata:
                        if key in _data:
                            _data[key].update(_tmpdata[key])
                        else:
                            _data[key] = _tmpdata[key]
                else:
                    for key in _tmpdata:
                        if key is None:
                            _data[scenario].update(_tmpdata[key])
                        else:
                            raise IOError(
                                "Cannot define a scenario within another scenario"
                            )
            else:
                _process_data(cmd, _model, _data[scenario], _default, Filename, Lineno)
    return True


def _process_table(cmd, _model, _data, _default, options=None):
    # print("TABLE %s" % cmd)
    #
    _options = {}
    _set = OrderedDict()
    _param = OrderedDict()
    _labels = []

    _cmd = cmd[1]
    _cmd_len = len(_cmd)
    name = None
    i = 0
    while i < _cmd_len:
        try:
            # print("CMD i=%s cmd=%s" % (i, _cmd[i:]))
            #
            # This should not be error prone, so we treat errors
            # with a general exception
            #

            #
            # Processing labels
            #
            if _cmd[i] == ':':
                i += 1
                while i < _cmd_len:
                    _labels.append(_cmd[i])
                    i += 1
                continue
            #
            # Processing options
            #
            name = _cmd[i]
            if i + 1 == _cmd_len:
                _param[name] = []
                _labels = ['Z']
                i += 1
                continue
            if _cmd[i + 1] == '=':
                if type(_cmd[i + 2]) is list:
                    _set[name] = _cmd[i + 2]
                else:
                    _options[name] = _cmd[i + 2]
                i += 3
                continue
            # This should be a parameter declaration
            if not type(_cmd[i + 1]) is tuple:
                raise IOError
            if i + 2 < _cmd_len and _cmd[i + 2] == '=':
                _param[name] = (_cmd[i + 1], _cmd[i + 3][0])
                i += 4
            else:
                _param[name] = _cmd[i + 1]
                i += 2
        except:
            raise IOError("Error parsing table options: %s" % name)

    # print("_options %s" % _options)
    # print("_set %s" % _set)
    # print("_param %s" % _param)
    # print("_labels %s" % _labels)
    #
    options = Bunch(**_options)
    for key in options:
        if not key in ['columns']:
            raise ValueError("Unknown table option '%s'" % key)
    #
    ncolumns = options.columns
    if ncolumns is None:
        ncolumns = len(_labels)
        if ncolumns == 0:
            if not (len(_set) == 1 and len(_set[_set.keys()[0]]) == 0):
                raise IOError(
                    "Must specify either the 'columns' option or column headers"
                )
            else:
                ncolumns = 1
    else:
        ncolumns = int(ncolumns)
    #
    data = cmd[2]
    Ldata = len(cmd[2])
    #
    cmap = {}
    if len(_labels) == 0:
        for i in range(ncolumns):
            cmap[i + 1] = i
        for label in _param:
            ndx = cmap[_param[label][1]]
            if ndx < 0 or ndx >= ncolumns:
                raise IOError("Bad column value %s for data %s" % (str(ndx), label))
            cmap[label] = ndx
            _param[label] = _param[label][0]
    else:
        i = 0
        for label in _labels:
            cmap[label] = i
            i += 1
    # print("CMAP %s" % cmap)
    #
    # print("_param %s" % _param)
    # print("_set %s" % _set)
    for sname in _set:
        # Creating set sname
        cols = _set[sname]
        tmp = []
        for col in cols:
            if not col in cmap:
                raise IOError(
                    "Unexpected table column '%s' for index set '%s'" % (col, sname)
                )
            tmp.append(cmap[col])
        if not sname in cmap:
            cmap[sname] = tmp
        cols = list(flatten_tuple(tmp))
        #
        _cmd = ['set', sname, ':=']
        i = 0
        while i < Ldata:
            row = []
            # print("COLS %s  NCOLS %d" % (cols, ncolumns))
            for col in cols:
                # print("Y %s %s" % (i, col))
                row.append(data[i + col])
            if len(row) > 1:
                _cmd.append(tuple(row))
            else:
                _cmd.append(row[0])
            i += ncolumns
        # print("_data %s" % _data)
        _process_set(_cmd, _model, _data)
    #
    # print("CMAP %s" % cmap)
    _i = 0
    if ncolumns == 0:
        raise IOError
    for vname in _param:
        _i += 1
        # create value vname
        cols = _param[vname]
        tmp = []
        for col in cols:
            # print("COL %s" % col)
            if not col in cmap:
                raise IOError(
                    "Unexpected table column '%s' for table value '%s'" % (col, vname)
                )
            tmp.append(cmap[col])
        # print("X %s %s" % (len(cols), tmp))
        cols = list(flatten_tuple(tmp))
        # print("X %s" % len(cols))
        # print("VNAME %s %s" % (vname, cmap[vname]))
        if vname in cmap:
            cols.append(cmap[vname])
        else:
            cols.append(ncolumns - 1 - (len(_param) - _i))
        # print("X %s" % len(cols))
        #
        _cmd = ['param', vname, ':=']
        i = 0
        while i < Ldata:
            # print("HERE %s %s %s" % (i, cols, ncolumns))
            for col in cols:
                _cmd.append(data[i + col])
            i += ncolumns
        # print("HERE %s" % _cmd)
        # print("_data %s" % _data)
        _process_param(_cmd, _model, _data, None, ncolumns=len(cols))


def _process_load(cmd, _model, _data, _default, options=None):
    # print("LOAD %s" % cmd)
    from pyomo.core import Set

    _cmd_len = len(cmd)
    _options = {}
    _options['filename'] = cmd[1]
    i = 2
    while cmd[i] != ':':
        _options[cmd[i]] = cmd[i + 2]
        i += 3
    i += 1
    _Index = (None, [])
    if type(cmd[i]) is tuple:
        _Index = (None, cmd[i])
        i += 1
    elif i + 1 < _cmd_len and cmd[i + 1] == '=':
        _Index = (cmd[i], cmd[i + 2])
        i += 3
    _smap = OrderedDict()
    while i < _cmd_len:
        if i + 2 < _cmd_len and cmd[i + 1] == '=':
            _smap[cmd[i + 2]] = cmd[i]
            i += 3
        else:
            _smap[cmd[i]] = cmd[i]
            i += 1

    if len(cmd) < 2:
        raise IOError("The 'load' command must specify a filename")

    options = Bunch(**_options)
    for key in options:
        if not key in [
            'range',
            'filename',
            'format',
            'using',
            'driver',
            'query',
            'table',
            'user',
            'password',
            'database',
        ]:
            raise ValueError("Unknown load option '%s'" % key)

    global Filename
    Filename = options.filename

    global Lineno
    Lineno = 0
    #
    # TODO: process mapping info
    #
    if options.using is None:
        tmp = options.filename.split(".")[-1]
        data = DataManagerFactory(tmp)
        if (data is None) or isinstance(data, UnknownDataManager):
            raise ApplicationError("Data manager '%s' is not available." % tmp)
    else:
        try:
            data = DataManagerFactory(options.using)
        except:
            data = None
        if (data is None) or isinstance(data, UnknownDataManager):
            raise ApplicationError(
                "Data manager '%s' is not available." % options.using
            )
    set_name = None
    #
    # Create symbol map
    #
    symb_map = _smap
    if len(symb_map) == 0:
        raise IOError(
            "Must specify at least one set or parameter name that will be loaded"
        )
    #
    # Process index data
    #
    _index = None
    index_name = _Index[0]
    _select = None
    #
    # Set the 'set name' based on the format
    #
    _set = None
    if options.format == 'set' or options.format == 'set_array':
        if len(_smap) != 1:
            raise IOError(
                "A single set name must be specified when using format '%s'"
                % options.format
            )
        set_name = list(_smap.keys())[0]
        _set = set_name
    #
    # Set the 'param name' based on the format
    #
    _param = None
    if (
        options.format == 'transposed_array'
        or options.format == 'array'
        or options.format == 'param'
    ):
        if len(_smap) != 1:
            raise IOError(
                "A single parameter name must be specified when using format '%s'"
                % options.format
            )
    if options.format in ('transposed_array', 'array', 'param', None):
        if _Index[0] is None:
            _index = None
        else:
            _index = _Index[0]
        _param = []
        _select = list(_Index[1])
        for key in _smap:
            _param.append(_smap[key])
            _select.append(key)
    if options.format in ('transposed_array', 'array'):
        _select = None

    # print "YYY", _param, options
    if (
        not _param is None
        and len(_param) == 1
        and not _model is None
        and isinstance(getattr(_model, _param[0]), Set)
    ):
        _select = None
        _set = _param[0]
        _param = None
        _index = None

    # print "SELECT", _param, _select
    #
    data.initialize(
        model=options.model,
        filename=options.filename,
        index=_index,
        index_name=index_name,
        param_name=symb_map,
        set=_set,
        param=_param,
        format=options.format,
        range=options.range,
        query=options.query,
        using=options.using,
        table=options.table,
        select=_select,
        user=options.user,
        password=options.password,
        database=options.database,
    )
    #
    data.open()
    try:
        data.read()
    except Exception:
        data.close()
        raise
    data.close()
    data.process(_model, _data, _default)


def _process_data(
    cmd,
    _model,
    _data,
    _default,
    Filename_,
    Lineno_=0,
    index=None,
    set=None,
    param=None,
    ncolumns=None,
):
    """
    Called by import_file() to (1) preprocess data and (2) call
    subroutines to process different types of data
    """
    # print("CMD %s" %cmd)
    global Lineno
    global Filename
    Lineno = Lineno_
    Filename = Filename_
    generate_debug_messages = is_debug_set(logger)
    if generate_debug_messages:
        logger.debug("DEBUG: _process_data (start) %s", cmd)
    if len(cmd) == 0:  # pragma:nocover
        raise ValueError("ERROR: Empty list passed to Model::_process_data")

    if cmd[0] == "data":
        return True

    if cmd[0] == "end":
        return False

    if cmd[0].startswith('set'):
        cmd = _preprocess_data(cmd)
        _process_set(cmd, _model, _data)

    elif cmd[0].startswith('param'):
        cmd = _preprocess_data(cmd)
        _process_param(
            cmd, _model, _data, _default, index=index, param=param, ncolumns=ncolumns
        )

    elif cmd[0] == 'include':
        cmd = _preprocess_data(cmd)
        _process_include(cmd, _model, _data, _default)

    elif cmd[0] == 'load':
        cmd = _preprocess_data(cmd)
        _process_load(cmd, _model, _data, _default)

    elif cmd[0] == 'table':
        cmd = [cmd[0], _preprocess_data(cmd[1]), _preprocess_data(cmd[2])]
        _process_table(cmd, _model, _data, _default)

    else:
        raise IOError("ERROR: Unknown data command: " + " ".join(cmd))

    return True
